QUERY__setup_system = "SELECT setup();"
QUERY__load_application_configuration = "SELECT * FROM application_configuration WHERE name = :configuration AND application = :application"
QUERY__load_application_configurations = "SELECT * FROM application_configuration"
QUERY__fetch_internal_admin = "SELECT * FROM check_admin"
QUERY__fetch_recent_passwords_with_ips = "SELECT * FROM fetch_recent_passwords_with_ips WHERE username = :enc_username AND encrypted_address = :encrypted_address;"
QUERY__fetch_recent_passwords = "SELECT * FROM fetch_recent_passwords WHERE username = :enc_username;"
QUERY__fetch_recent_passwords_with_account = "SELECT * FROM fetch_recent_passwords WHERE user_id = :user_id;"
QUERY__create_account = "SELECT create_account(:username, :application, :configuration, :password)"
QUERY__fetch_matching_recent_passwords_with_account = "SELECT * FROM fetch_recent_passwords WHERE user_id = :user_id AND password_id = :password;"
QUERY__add_password = "INSERT INTO account_password (account, password_hash) VALUES (:account, :password_hash) RETURNING id;"
QUERY__add_password_cache = "UPDATE account SET password = :password WHERE user_id = :account"
QUERY__add_or_update_ip = "INSERT INTO account_ip (account, encrypted_address) VALUES (:account, :encrypted_address) ON CONFLICT ON CONSTRAINT user_ip_unq DO UPDATE SET most_recent_login = current_timestamp RETURNING id as address"
QUERY__setup_jaaql_role = "SELECT setup_jaaql_role();"
QUERY__attach_account = "SELECT attach_account(:user_id, :enc_username);"
QUERY__fetch_user_from_username = "SELECT user_id FROM account WHERE username = :username"
QUERY__fetch_user_record_from_username = "SELECT * FROM account WHERE username = :username"
QUERY__fetch_application_schemas = "SELECT * FROM application_schemas"
QUERY__postgres_version = "SELECT version() as version;"
QUERY__drop_email_account = "SELECT drop_email_account(:name)"
QUERY__drop_database = "SELECT drop_database(:name)"
QUERY__fetch_public_account_credentials = "SELECT username, public_password as password FROM account WHERE application = :application AND configuration = :configuration"
QUERY__fetch_email_template_with_app_url = "SELECT et.*, ac.artifact_base_uri, ac.url, etc.account, etc.override_send_name, etc.account FROM email_template et INNER JOIN application_configuration ac ON ac.application = et.application INNER JOIN email_template_configuration etc ON ac.application = etc.application AND ac.name = etc.configuration AND et.name = etc.template WHERE ac.name = :configuration and et.name = :name AND et.application = :application"
QUERY__fetch_table_columns = "SELECT column_name, is_primary FROM jaaql__table_cols_marked_primary WHERE table_name = :table_name"
QUERY__sign_up_count = "SELECT COUNT(*) as count FROM sign_up WHERE account = :user_id AND (created + interval '24 hour') > current_timestamp"
QUERY__sign_up_insert = "INSERT INTO sign_up (account, email_template, data_lookup_json, invite_code, configuration, application) VALUES (:the_user, :template, :data_lookup_json, :invite_code, :configuration, :application) RETURNING key_a as invite_key, key_b as invite_poll_key, invite_code as invite_code"
QUERY__sign_up_poll = "SELECT key_a as invite_key, invite_code, activated, su.created, su.account, su.closed, us.username as email, su.code_expiry_ms, su.used_key_a, su.code_attempts FROM sign_up su INNER JOIN account us ON us.user_id = su.account WHERE (key_a = :invite_or_poll_key or key_b = :invite_or_poll_key) AND extract(epoch FROM (current_timestamp - su.created)) * 1000 < expiry_ms"
QUERY__sign_up_upd = "UPDATE sign_up SET activated = TRUE WHERE key_a = :invite_key"
QUERY__sign_up_upd_used = "UPDATE sign_up SET used_key_a = TRUE WHERE key_a = :invite_key"
QUERY__sign_up_increment_attempts = "UPDATE sign_up SET code_attempts = code_attempts + 1 WHERE key_a = :invite_key"
QUERY__sign_up_fetch = "SELECT su.key_a as invite_key, su.activated, su.account, su.closed, su.data_lookup_json, us.username as email, su.email_template as template, su.data_lookup_json, su.configuration, su.application FROM sign_up su INNER JOIN account us on us.user_id = su.account WHERE ((su.key_a = :invite_or_poll_key) or (su.activated and su.key_b = :invite_or_poll_key)) AND extract(epoch FROM (current_timestamp - su.created)) * 1000 < su.expiry_ms AND su.closed is null"
QUERY__sign_up_close = "UPDATE sign_up SET closed = current_timestamp WHERE key_a = :invite_key"
QUERY__reset_count = "SELECT COUNT(*) as count FROM reset_password WHERE the_user = :the_user AND (created + interval '24 hour') > current_timestamp"
QUERY__fake_reset_count = "SELECT COUNT(*) as count FROM fake_reset_password WHERE email = :email AND (created + interval '24 hour') > current_timestamp"
QUERY__reset_insert = "INSERT INTO reset_password (the_user, email_template, data_lookup_json, reset_code, configuration, application) VALUES (:the_user, :template, :data_lookup_json, :reset_code, :configuration, :application) RETURNING key_a as reset_key, key_b as reset_poll_key, reset_code as reset_code"
QUERY__fake_reset_insert = "INSERT INTO fake_reset_password (email) VALUES (:email) RETURNING key_b as reset_poll_key"
QUERY__reset_poll = "SELECT key_a as reset_key, reset_code, activated, rp.created, rp.the_user, rp.closed, us.username as email, rp.code_expiry_ms, rp.used_key_a, rp.code_attempts FROM reset_password rp INNER JOIN account us ON us.user_id = rp.the_user WHERE (key_a = :reset_or_poll_key or key_b = :reset_or_poll_key) AND extract(epoch FROM (current_timestamp - rp.created)) * 1000 < expiry_ms"
QUERY__fake_reset_poll = "SELECT gen_random_uuid() as reset_key, rp.created, rp.email, rp.code_attempts, FALSE as activated, rp.code_expiry_ms, gen_random_uuid() as reset_code FROM fake_reset_password rp WHERE key_b = :reset_or_poll_key AND extract(epoch FROM (current_timestamp - rp.created)) * 1000 < expiry_ms"
QUERY__reset_upd = "UPDATE reset_password SET activated = TRUE WHERE key_a = :reset_key"
QUERY__reset_upd_used = "UPDATE reset_password SET used_key_a = TRUE WHERE key_a = :reset_key"
QUERY__fake_reset_increment_attempts = "UPDATE fake_reset_password SET code_attempts = code_attempts + 1 WHERE key_b = :reset_key"
QUERY__reset_increment_attempts = "UPDATE reset_password SET code_attempts = code_attempts + 1 WHERE key_a = :reset_key"
QUERY__reset_fetch = "SELECT rp.key_a as reset_key, rp.activated, rp.the_user, rp.closed, us.username as email, rp.email_template as template, rp.data_lookup_json, rp.configuration, rp.application FROM reset_password rp INNER JOIN account us on us.user_id = rp.the_user WHERE ((rp.key_a = :reset_or_poll_key) or (rp.activated and rp.key_b = :reset_or_poll_key)) AND extract(epoch FROM (current_timestamp - rp.created)) * 1000 < rp.expiry_ms AND rp.closed is null"
QUERY__reset_close = "UPDATE reset_password SET closed = current_timestamp WHERE key_a = :reset_key or key_b = :reset_key"
QUERY__ins_rendered_document = "INSERT INTO rendered_document (encrypted_parameters, encrypted_access_token, document, create_file, application, configuration) VALUES (:parameters, :oauth_token, :name, :create_file, :application, :configuration) RETURNING document_id"
QUERY__purge_rendered_document = "DELETE FROM rendered_document WHERE completed is not null and document_id = :document_id RETURNING content"
QUERY__fetch_rendered_document = "SELECT rd.document_id, able.render_as, rd.filename, rd.create_file, rd.completed, rd.encrypted_access_token as oauth_token FROM rendered_document rd INNER JOIN renderable_document able ON rd.document = able.name WHERE rd.document_id = :document_id"
QUERY__log_ins = "INSERT INTO log (the_user, occurred, duration_ms, encrypted_exception, encrypted_input, ip, status, endpoint) VALUES (:user_id, :occurred, :duration_ms, :exception, :input, :ip, :status, :endpoint)"



from enum import IntEnum

__all__ = ['CustomHTTPStatus']


class CustomHTTPStatus(IntEnum):

    def __new__(cls, value, phrase, description=''):
        obj = int.__new__(cls, value)
        obj._value_ = value

        obj.phrase = phrase
        obj.description = description
        return obj

    DATABASE_NO_EXIST = 460, "Database doesn't exist", "The database doesn't exist on the given node"



import sys
import subprocess
import glob
import unittest
from jaaql.test.components.component_utils import wait_for_service


def is_test_file(file):
    return file.split(".")[-2] not in ["base_component", "__init__", "component_utils"]


def run_component_tests():
    test_files = [
        ".".join(test_file.replace("\\", ".").replace("/", ".").split(".")[:-1])
        for test_file in glob.glob("components/*.py", recursive=True)
        if is_test_file(test_file.replace("\\", ".").replace("/", "."))
    ]
    test_files.sort()

    suites = [
        unittest.defaultTestLoader.loadTestsFromName(test_file)
        for test_file in test_files
    ]
    test_suite = unittest.TestSuite(suites)
    runner = unittest.TextTestRunner(verbosity=2)  # For local debugging
    result = runner.run(test_suite)

    if len(result.errors) != 0 or len(result.failures) != 0:
        print("Component test failure!", file=sys.stderr)
    else:
        print("Component tests all passed")

    if not sys.platform.lower().startswith('win'):
        pid = open("app.pid", "r").read()
        subprocess.call("kill -HUP " + pid, shell=True)  # Kill gunicorn so coverage report is generated


if __name__ == "__main__":
    wait_for_service()

    run_component_tests()

    if not sys.platform.lower().startswith('win'):
        app_pid = open("app.pid", "r").read()
        subprocess.call("kill -HUP " + app_pid, shell=True)  # Kill gunicorn so coverage report is generated



from jaaql.patch import monkey_patch

monkey_patch()

from jaaql.services.script_install import bootup
import os
from jaaql.constants import ENVIRON__vault_key, ENVIRON__local_install

bootup(os.environ.get(ENVIRON__vault_key), True, os.environ.get(ENVIRON__local_install) == "TRUE")



import os

from .base_component import *
import imaplib
import email
import ssl

MAX_EMAIL_TIME_MS = 300000
EMAIL_HOST_URL = "web119.shared.hosting-login.net"
EMAIL_USERNAME = "jaaql-component-receive@sqmi.nl"
EMAIL_ENV_PASSWORD = "COMPONENT_EMAIL_PASSWORD"
EMAIL_IMAP_PORT = 993
EMAIL_FROM = "jaaql-component-send@sqmi.nl"


class SignupComponent(BaseComponent):

    def test_signup(self):
        self.create_email_template()

    def create_email_template(self):
        pass
        # requests.post(BASE_INTERNAL_URL + "/internal/emails/templates", json={
        #     "name": "signup",
        #     "account": "Notification account",
        #     "description": "The signup email template",
        #     "app_relative_path": "my_template",
        #     "subject": "Welcome to JAAQL",
        #     "allow_signup": True,
        #     "allow_confirm_signup_attempt": False,
        #     "data_validation_table": "ctp_data_validation_table"
        # })

    def perform_signup_tests(self):
        pass
#         auth_token = self.get_jaaql_auth_header()
#
#         replacement_data = "replaced at " + str(datetime.now())
#         resp = requests.post(BASE_URL + "/ctp/send_email", json={
#             "email_data": replacement_data
#         }, headers={HEADER_AUTH: auth_token})
#
#         self.assertEqual(HTTPStatus.OK, resp.status_code, "Response ok")
#
#         expected_subject = "CTP Test Email"
#         expected_body = """<html lang="en">
#     <head><title>CTP Test Email</title></head>
#     <body>
#         This is the CTP test email with data %s
#     </body>
# </html>""" % replacement_data
#         expected_body.replace("\r\n", "\n").replace("\n", "\r\n")  # Not a mistake. Double prevents \r\r\n
#         expected_attachment = "This is the content of the email attachment"
#         expected_attachment_filename = "email_attachment.txt"
#
#         actual_subject, actual_body, actual_attachment, actual_attachment_filename = self.read_email()
#
#         self.assertEqual(expected_subject, actual_subject)
#         self.assertEqual(expected_body, actual_body)
#         self.assertEqual(expected_attachment, actual_attachment)
#         self.assertEqual(expected_attachment_filename, actual_attachment_filename)

    def fetch_imap_conn(self, context: ssl.SSLContext):
        conn = imaplib.IMAP4_SSL(EMAIL_IMAP_URL, EMAIL_IMAP_PORT, ssl_context=context)
        conn.login(EMAIL_USERNAME, os.environ[EMAIL_ENV_PASSWORD])
        conn.select(readonly=0)
        return conn

    def read_email(self):
        context = ssl.SSLContext()
        conn = self.fetch_imap_conn(context)
        start_time = datetime.now()

        email_subject = None
        email_body = None
        email_attachment = None
        email_attachment_filename = None

        while time_delta_ms(start_time, datetime.now()) < MAX_EMAIL_TIME_MS:
            try:
                retcode, messages = conn.search(None, '(UNSEEN)')
                for msg_id in messages[0].split():
                    typ, data = conn.fetch(msg_id, '(RFC822)')
                    msg = email.message_from_bytes(data[0][1])
                    conn.store(msg_id, '+FLAGS', '\\Seen')
                    from_email = msg['From']

                    if '<' in from_email:
                        from_email = from_email.split("<")[1].split(">")[0]

                    if from_email == EMAIL_FROM:
                        email_subject = msg['subject']
                        for part in msg.walk():
                            if part.get_filename() is None:
                                if part.get_content_type() == "text/plain" and email_body is None:
                                    email_body = part.get_payload(decode=True).decode("ASCII")
                                elif part.get_content_type() == "text/html":
                                    email_body = part.get_payload(decode=True).decode("ASCII")
                            else:
                                email_attachment = part.get_payload(decode=True).decode("ASCII")
                                email_attachment_filename = part.get_filename()
                        break
                if email_body is not None:
                    break
            except imaplib.IMAP4.error as ex:
                conn.logout()
                conn = self.fetch_imap_conn(context)
            time.sleep(5)

        return email_subject, email_body, email_attachment, email_attachment_filename



from typing import List

OPTION__short_max = 2
OPTION__required = "required"
OPTION__optional = "optional"
OPTION__flag = "flag    "
OPTION__parameter = "parameter"
OPTION__short_id = "-"
OPTION__long_id = "--"

ERR__missing_required_options = "ERROR: Missing required options!"
ERR__missing_option = "ERROR: Missing option '%s'!"
ERR__option_short_too_long = "Option %d has short description longer than %d!"
ERR__expected_option = "Expected option, instead found '%s'!"
ERR__no_matching_option = "Could not find option with identifier '%s'!"
ERR__expected_argument = "Expected argument for option '%s'!"
ERR__option_already_exists = "Option '%s' supplied more than once!"
ERR__missing_argument = "Missing argument for option '%s'!"


class Option:

    def __init__(self, short: str, long: str, description: str, is_flag: bool, required=True):
        self.short = short
        self.long = long
        self.required = required
        self.description = description
        self.is_flag = is_flag


OPT_KEY__vault_key = "vault-key"
OPT_KEY__email_credentials = "email-credentials"
OPT_KEY__help = "help"
OPT_KEY__profiling = "profiling"
OPT_KEY__local_install = "local_install"
OPT_KEY__canned_queries = "canned_queries"

DEFAULT_OPTIONS: List[Option] = [
    Option(
        short="h",
        long=OPT_KEY__help,
        required=False,
        description="Outputs the help directory",
        is_flag=True
    ),
    Option(
        short="v",
        long=OPT_KEY__vault_key,
        required=False,
        description="Encrypts the jaaql vault which stores internal sensitive information",
        is_flag=False
    ),
    Option(
        short="e",
        long=OPT_KEY__email_credentials,
        required=False,
        description="Provides SMTP/IMAP email credentials into the JAAQL server. Base64 encoded json dict of the format"
        " { \"my_account_name\": \"my_account_password\" }",
        is_flag=False
    ),
    Option(
        short="p",
        long=OPT_KEY__profiling,
        required=False,
        description="Will now output profiling information about the request",
        is_flag=True
    ),
    Option(
        short="i",
        long=OPT_KEY__local_install,
        required=False,
        description="Will now install on startup",
        is_flag=True
    ),
    Option(
        short="c",
        long=OPT_KEY__canned_queries,
        required=False,
        description="Will allow only pre-selected queries for users without DBA access",
        is_flag=True
    )
]


def parse_options(args: List[str], do_print_help: bool = True, additional_options: List[Option] = None, title: str = None,
                  use_options: [Option] = None):
    if use_options is None:
        use_options = DEFAULT_OPTIONS

    args = args[1:]

    if additional_options is not None:
        use_options += additional_options

    if len(args) == 0 and do_print_help:
        print_help(title, use_options)
        print()
        print(ERR__missing_required_options)
        exit(0)

    found_options = {}

    cur_option = None
    for i, cur_arg in zip(range(len(args)), args):
        filled_arg = False

        if cur_option is None:
            matching = None

            if cur_arg.startswith(OPTION__long_id):
                matching = [option for option in use_options if option.long == cur_arg[len(OPTION__long_id):]]
            elif cur_arg.startswith(OPTION__short_id):
                matching = [option for option in use_options if option.short == cur_arg[len(OPTION__short_id):]]

            if matching is None:
                raise Exception(ERR__expected_option % cur_arg)
            if len(matching) == 0:
                raise Exception(ERR__no_matching_option % cur_arg)

            cur_option = matching[0]
            if cur_option.long in found_options:
                raise Exception(ERR__option_already_exists % cur_option.long)

            found_options[cur_option.long] = None
        else:
            found_options[cur_option.long] = cur_arg
            filled_arg = True

        if cur_option.is_flag or filled_arg:
            cur_option = None

    if cur_option is not None and not cur_option.is_flag:
        raise Exception(ERR__missing_argument % cur_option.long)

    if OPT_KEY__help in found_options:
        print_help(title, use_options)
        exit(0)

    missing_options = [option for option in use_options if option.required and option.long not in found_options]
    if len(missing_options) != 0:
        raise Exception(ERR__missing_option % missing_options[0].long)

    return found_options


DEFAULT_TITLE = "JAAQL middleware system"


def print_help(title: str = None, use_options: [Option] = None):
    if use_options is None:
        use_options = DEFAULT_OPTIONS

    if title is None:
        title = DEFAULT_TITLE
    print("The %s, written in python" % title)
    print()
    print("OPTIONS:")

    max_long_spaces = max([len(option.long) for option in use_options])

    for option in use_options:
        if len(option.short) > OPTION__short_max:
            raise Exception(ERR__option_short_too_long % (option.short, OPTION__short_max))

        short_spaces = "".join([" "] * OPTION__short_max)[:OPTION__short_max - len(option.short)] + "\t"
        long_spaces = "".join([" "] * max_long_spaces)[:max_long_spaces - len(option.long)] + "\t"

        required = OPTION__required if option.required else OPTION__optional
        required += "\t"

        option_match = option.long + long_spaces + "-" + option.short + short_spaces
        is_flag = OPTION__flag if option.is_flag else OPTION__parameter
        is_flag += "\t"

        print("\t--" + option_match + required + is_flag + option.description)



import os

from .base_component import *
from jaaql.constants import *
from base64 import b64decode as b64d
from pyzbar.pyzbar import decode, ZBarSymbol
import pyotp
from PIL import Image
from io import BytesIO

KEY_INSTALL_KEY = "install_key"
PATH_INSTALL = "/install"
PATH_UNINSTALL = "/uninstall"

ENVIRON__fast_jaaql_component_tests = "FAST_TESTS"


class InstallComponent(BaseComponent):

    def test_install(self):
        if not bool(os.environ.get(ENVIRON__fast_jaaql_component_tests, False)):
            self.run_test_install()
            self.run_test_install_mfa()
        else:
            print("Skipping full install tests as fast tests activated")
        self.run_test_install_superjaaql()

    def get_uninstall_data(self, db_user_password: str, uninstall_key: str):
        return {
            KEY__db_super_user_password: db_user_password,
            KEY__uninstall_key: uninstall_key
        }

    def run_test_install(self):
        resp = self.get_jaaql_auth_header(run_test=False)
        self.assertEqual(HTTPStatus.INTERNAL_SERVER_ERROR, resp.status_code, "Not yet installed")

        resp = requests.post(BASE_INTERNAL_URL + PATH_UNINSTALL, json=self.get_uninstall_data(os.environ[PG_ENV__password],
                                                                                              self.fetch_uninstall_key()))
        self.assertEqual(HTTPStatus.UNPROCESSABLE_ENTITY, resp.status_code, "Not yet installed, uninstall")

        install_data = {
            "password": PASSWORD_JAAQL,
            "use_mfa": False,
            KEY_INSTALL_KEY: "wrong_install_key"
        }
        resp = requests.post(BASE_INTERNAL_URL + PATH_INSTALL, json=install_data)
        self.assertEqual(HTTPStatus.UNAUTHORIZED, resp.status_code, "Wrong install key")

        install_data = {
            "password": PASSWORD_JAAQL,
            "use_mfa": False,
            KEY_INSTALL_KEY: self.fetch_install_key(),
            "allow_uninstall": True
        }

        resp = requests.post(BASE_INTERNAL_URL + PATH_INSTALL, json=install_data)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Correct install key")
        self.assertEqual(None, resp.json()[KEY__jaaql_otp_uri], "Otp URI missing")
        self.assertEqual(None, resp.json()[KEY__jaaql_otp_qr], "Otp QR missing")
        self.assertEqual(None, resp.json()[KEY__superjaaql_otp_uri], "Superjaaql OTP URI missing")
        self.assertEqual(None, resp.json()[KEY__superjaaql_otp_qr], "Superjaaql OTP QR missing")

        time.sleep(20)
        wait_for_service()

        resp = requests.post(BASE_INTERNAL_URL + PATH_INSTALL, json=install_data)
        self.assertEqual(HTTPStatus.UNPROCESSABLE_ENTITY, resp.status_code, "Already installed")

        resp = self.get_jaaql_auth_header(run_test=False)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "JAAQL can login")
        resp = self.get_jaaql_auth_header(password="thewrongpassword", run_test=False)
        self.assertEqual(HTTPStatus.UNAUTHORIZED, resp.status_code, "Jaaql can't login with the incorrect password")
        resp = self.get_jaaql_auth_header(username="superjaaql", password=PASSWORD_SUPERJAAQL, run_test=False)
        self.assertEqual(HTTPStatus.UNAUTHORIZED, resp.status_code, "Superjaaql can't login")

        resp = requests.post(BASE_INTERNAL_URL + PATH_UNINSTALL, json=self.get_uninstall_data("wrong password", self.fetch_uninstall_key()))
        self.assertEqual(HTTPStatus.UNPROCESSABLE_ENTITY, resp.status_code, "Uninstall was unsuccessful, wrong db password")
        resp = requests.post(BASE_INTERNAL_URL + PATH_UNINSTALL, json=self.get_uninstall_data(os.environ[PG_ENV__password], "wrong uninstall key"))
        self.assertEqual(HTTPStatus.UNAUTHORIZED, resp.status_code, "Uninstall was unsuccessful, wrong uninstall key")
        resp = requests.post(BASE_INTERNAL_URL + PATH_UNINSTALL, json=self.get_uninstall_data(os.environ[PG_ENV__password],
                                                                                              self.fetch_uninstall_key()))
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Uninstall was successful")

        time.sleep(20)
        wait_for_service()

        resp = requests.post(BASE_INTERNAL_URL + PATH_UNINSTALL, json=self.get_uninstall_data(os.environ[PG_ENV__password],
                                                                                              self.fetch_uninstall_key()))
        self.assertEqual(HTTPStatus.UNPROCESSABLE_ENTITY, resp.status_code, "Not yet installed, after uninstall")

    def run_test_install_mfa(self):
        install_data = {
            "password": PASSWORD_JAAQL,
            "use_mfa": True,
            KEY_INSTALL_KEY: self.fetch_install_key(),
            "allow_uninstall": True
        }

        resp = requests.post(BASE_INTERNAL_URL + PATH_INSTALL, json=install_data)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Correct install key")
        self.assertEqual(True, resp.json()[KEY__jaaql_otp_uri] is not None, "Otp URI present")
        self.assertEqual(True, resp.json()[KEY__jaaql_otp_qr] is not None, "Otp QR present")
        self.assertEqual(None, resp.json()[KEY__superjaaql_otp_uri], "Superjaaql OTP URI missing")
        self.assertEqual(None, resp.json()[KEY__superjaaql_otp_qr], "Superjaaql OTP QR missing")
        otp_uri = resp.json()[KEY__jaaql_otp_uri]

        time.sleep(20)
        wait_for_service()

        self.assertEqual(otp_uri, self.decode_b64_qr(resp.json()[KEY__jaaql_otp_qr]), "QR decodes jaaql")
        resp = self.get_jaaql_auth_header(run_test=False)
        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_code, "Pre auth check")
        resp_app = requests.get(BASE_URL + "/applications", headers={HEADER_AUTH: resp.json()})
        self.assertEqual(HTTPStatus.UNAUTHORIZED, resp_app.status_code, "Pre auth denied")
        resp_mfa = requests.post(BASE_URL + "/oauth/token", json={"mfa_key": "123456", "pre_auth_key": resp.json()})
        self.assertEqual(HTTPStatus.UNAUTHORIZED, resp_mfa.status_code, "Wrong MFA key")
        gen_mfa_key = self.gen_mfa_key(otp_uri)
        resp_mfa = requests.post(BASE_URL + "/oauth/token", json={"mfa_key": gen_mfa_key, "pre_auth_key": resp.json()})
        self.assertEqual(HTTPStatus.OK, resp_mfa.status_code, "Correct MFA key")
        resp = self.get_jaaql_auth_header(run_test=False)
        resp = requests.post(BASE_URL + "/oauth/token", json={"mfa_key": gen_mfa_key, "pre_auth_key": resp.json()})
        self.assertEqual(HTTPStatus.UNAUTHORIZED, resp.status_code, "Cannot reuse mfa key")

        requests.post(BASE_INTERNAL_URL + PATH_UNINSTALL, json=self.get_uninstall_data(os.environ[PG_ENV__password], self.fetch_uninstall_key()))
        time.sleep(20)
        wait_for_service()

        install_data = {
            "password": PASSWORD_JAAQL,
            "superjaaql_password": PASSWORD_SUPERJAAQL,
            "use_mfa": True,
            KEY_INSTALL_KEY: self.fetch_install_key(),
            "allow_uninstall": True
        }

        resp = requests.post(BASE_INTERNAL_URL + PATH_INSTALL, json=install_data)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Correct install key")
        self.assertEqual(True, resp.json()[KEY__jaaql_otp_uri] is not None, "Otp URI present")
        self.assertEqual(True, resp.json()[KEY__jaaql_otp_qr] is not None, "Otp QR present")
        self.assertEqual(True, resp.json()[KEY__superjaaql_otp_uri] is not None, "Superjaaql OTP URI present")
        self.assertEqual(True, resp.json()[KEY__superjaaql_otp_qr] is not None, "Superjaaql OTP QR present")
        otp_uri = resp.json()[KEY__superjaaql_otp_uri]

        time.sleep(20)
        wait_for_service()

        self.assertEqual(otp_uri, self.decode_b64_qr(resp.json()[KEY__superjaaql_otp_qr]), "QR decodes superjaaql")
        resp = self.get_jaaql_auth_header(username="superjaaql", password=PASSWORD_SUPERJAAQL, run_test=False)
        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_code, "Pre auth check")
        gen_mfa_key = self.gen_mfa_key(otp_uri)
        resp_mfa = requests.post(BASE_URL + "/oauth/token", json={"mfa_key": gen_mfa_key, "pre_auth_key": resp.json()})
        self.assertEqual(HTTPStatus.OK, resp_mfa.status_code, "Correct MFA key")

        requests.post(BASE_INTERNAL_URL + PATH_UNINSTALL, json=self.get_uninstall_data(os.environ[PG_ENV__password], self.fetch_uninstall_key()))
        time.sleep(20)
        wait_for_service()

    def run_test_install_superjaaql(self):
        install_data = {
            "password": PASSWORD_JAAQL,
            "superjaaql_password": PASSWORD_SUPERJAAQL,
            "use_mfa": False,
            KEY_INSTALL_KEY: self.fetch_install_key()
        }

        resp = requests.post(BASE_INTERNAL_URL + PATH_INSTALL, json=install_data)
        self.assertEqual(None, resp.json()[KEY__jaaql_otp_uri], "Otp URI missing")
        self.assertEqual(None, resp.json()[KEY__jaaql_otp_qr], "Otp QR missing")
        self.assertEqual(None, resp.json()[KEY__superjaaql_otp_uri], "Superjaaql OTP URI missing")
        self.assertEqual(None, resp.json()[KEY__superjaaql_otp_qr], "Superjaaql OTP QR missing")

        time.sleep(20)
        wait_for_service()

        resp = self.get_jaaql_auth_header(run_test=False)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "JAAQL can login")
        resp = self.get_jaaql_auth_header(password="thewrongpassword", run_test=False)
        self.assertEqual(HTTPStatus.UNAUTHORIZED, resp.status_code, "Jaaql can't login with the incorrect password")
        resp = self.get_jaaql_auth_header(username="superjaaql", password=PASSWORD_SUPERJAAQL, run_test=False)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Superjaaql can login")
        resp = self.get_jaaql_auth_header(username="superjaaql", password="wrongsuperjaaqlpassword", run_test=False)
        self.assertEqual(HTTPStatus.UNAUTHORIZED, resp.status_code, "Superjaaql can't login with the incorrect password")

        resp = requests.post(BASE_INTERNAL_URL + PATH_UNINSTALL, json=self.get_uninstall_data(os.environ[PG_ENV__password],
                                                                                              self.fetch_uninstall_key()))
        self.assertEqual(HTTPStatus.UNPROCESSABLE_ENTITY, resp.status_code, "Uninstall not allowed")

    def fetch_install_key(self):
        install_key = None
        with open("log/gunicorn.log", "r") as log_handle:
            install_key = log_handle.read().split("INSTALL KEY: ")[-1].split("\n")[0].strip()
        return install_key

    def fetch_uninstall_key(self):
        uninstall_key = ""
        with open("log/gunicorn.log", "r") as log_handle:
            try:
                uninstall_key = log_handle.read().split("UNINSTALL KEY: ")[-1].split("\n")[0].strip()
            except:
                pass
        return uninstall_key

    def decode_b64_qr(self, b64_qr: str):
        qr = b64d(b64_qr.split(HTML__base64_png)[1].encode("ASCII"))
        qr = Image.open(BytesIO(qr))
        return decode(qr, symbols=[ZBarSymbol.QRCODE])[0].data.decode("ASCII")

    def gen_mfa_key(self, otp_uri: str):
        totp_iv = otp_uri.split("secret=")[1].split("&")[0]
        return pyotp.TOTP(totp_iv).now()



import unittest
import inspect
from http import HTTPStatus
from .component_utils import *

BASE_INTERNAL_URL = BASE_URL + "/internal"

PASSWORD_JAAQL = "Abcdefg456"
PASSWORD_SUPERJAAQL = "Hijklmn123"

HEADER_AUTH = "Authentication-Token"


class BaseComponent(unittest.TestCase):

    def get_jaaql_auth_header(self, username: str = "jaaql", password: str = PASSWORD_JAAQL, run_test: bool = True):
        res = requests.post(BASE_URL + "/oauth/token", json={
            "username": username,
            "password": password
        })

        if run_test:
            self.assertEqual(HTTPStatus.OK, res.status_code, "Expecting auth token")
            return res.json()
        else:
            return res

    def assertEqual(self, first, second, msg=None):
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        print("\033[36m\033[1mASSERTION:\033[0m \033[33m\033[1m" + calframe[1][3] + "\033[0m " + msg)
        super().assertEqual(first, second, msg)



import uuid

from psycopg import OperationalError
from psycopg_pool import ConnectionPool
import queue
from psycopg.errors import ProgrammingError, InvalidParameterValue, UndefinedFunction
import threading
import traceback
from jaaql.constants import ERR__invalid_token

from jaaql.db.db_interface import DBInterface, ECHO__none, CHAR__newline
from jaaql.exceptions.http_status_exception import *
from jaaql.exceptions.custom_http_status import CustomHTTPStatus
from jaaql.constants import KEY__database

ERR__connect_db = "Could not create connection to database!"

PGCONN__min_conns = 5
PGCONN__max_conns = 10

TIMEOUT = 2.5

ERR__invalid_role = "Role not allowed, invalid format!"
ERR__must_use_canned_query = "Must use canned query as you are not an admin!"

QUERY__dba_query = "SELECT pg_has_role(datdba::regrole, 'MEMBER') FROM pg_database WHERE datname = %(database)s;"


class DBPGInterface(DBInterface):

    HOST_POOLS = {}
    HOST_POOLS_QUEUES = {}

    @staticmethod
    def put_conn_threaded(username: str, db_name: str, the_queue: queue.Queue):
        while True:
            try:
                conn, do_reset = the_queue.get()
                if do_reset:
                    with conn.cursor() as cursor:
                        cursor.execute("RESET ROLE;")
                        did_close = False
                        if hasattr(conn, "jaaql_reset_key"):
                            try:
                                cursor.execute("SELECT jaaql__reset_session_authorization('" + str(conn.jaaql_reset_key) + "');")
                            except:
                                did_close = True
                                conn.close()
                        if not did_close:
                            cursor.execute("RESET ALL;")
                            conn.commit()
                DBPGInterface.HOST_POOLS[username][db_name].putconn(conn)
            except Exception:
                pass  # Ignore, pool has likely been wiped

    def __init__(self, config, host: str, port: int, db_name: str, username: str, role: str = None, password: str = None, sub_role: str = None):
        super().__init__(config, host, username)

        self.role = role
        self.sub_role = sub_role
        if sub_role is not None:
            if len([ch for ch in sub_role if not ch.isalnum() and ch not in ['_', '-']]) != 0:
                raise HttpStatusException(ERR__invalid_role, HTTPStatus.UNAUTHORIZED)

        self.output_query_exceptions = config["DEBUG"]["output_query_exceptions"].lower() == "true"
        self.username = username
        self.db_name = db_name

        if self.username not in DBPGInterface.HOST_POOLS:
            DBPGInterface.HOST_POOLS[self.username] = {}
            DBPGInterface.HOST_POOLS_QUEUES[self.username] = {}

        user_pool = DBPGInterface.HOST_POOLS[self.username]

        self.conn_str = None
        the_pool = None

        if password is not None:
            try:
                conn_str = "user=" + username + " password=" + password + " dbname=" + db_name
                # Important we don't list the host as this will force a unix socket
                if host is not None and host not in ['localhost', '127.0.0.1']:
                    conn_str += " host=" + host

                if str(port) != "5432":
                    conn_str += " port=" + str(port)

                self.conn_str = conn_str

                if self.db_name not in user_pool:
                    the_pool = ConnectionPool(conn_str, min_size=PGCONN__min_conns, max_size=PGCONN__max_conns, max_lifetime=60 * 30)
                    the_pool.getconn(timeout=TIMEOUT)
                    user_pool[self.db_name] = the_pool
                    the_queue = queue.Queue()
                    DBPGInterface.HOST_POOLS_QUEUES[self.username][self.db_name] = the_queue
                    threading.Thread(target=DBPGInterface.put_conn_threaded, args=[self.username, self.db_name, the_queue], daemon=True).start()
            except OperationalError as ex:
                if the_pool is not None:
                    the_pool.close()
                if "does not exist" in str(ex).split("\"")[-1] or "couldn't get a connection after" in str(ex):
                    raise HttpStatusException("Database \"" + self.db_name + "\" does not exist",
                                              CustomHTTPStatus.DATABASE_NO_EXIST)
                else:
                    raise HttpStatusException(str(ex))

        self.pg_pool = user_pool[self.db_name]

    def _get_conn(self, allow_operational: bool = True):
        conn = self.pg_pool.getconn(timeout=TIMEOUT)
        conn.jaaql_reset_key = str(uuid.uuid4())
        if self.role is not None or self.sub_role is not None:
            try:
                with conn.cursor() as cursor:
                    if self.role is not None:
                        try:
                            cursor.execute("SELECT jaaql__set_session_authorization('" + self.role + "', '" + conn.jaaql_reset_key + "');")
                        except UndefinedFunction:
                            raise HttpStatusException("Database '%s' has not been configured for usage with JAAQL. Please ask the dba to run 'configure_database_for_use_with_jaaql' from the jaaql database" % self.db_name)
                    if self.sub_role is not None:
                        cursor.execute("SET ROLE \"" + self.sub_role + "\"")
            except InvalidParameterValue:
                raise HttpStatusException(ERR__invalid_token, HTTPStatus.UNAUTHORIZED)
            except OperationalError as ex:
                if allow_operational:
                    self.pg_pool.close()
                    DBPGInterface.HOST_POOLS[self.username][self.db_name] = ConnectionPool(self.conn_str, min_size=PGCONN__min_conns,
                                                                                           max_size=PGCONN__max_conns, max_lifetime=60 * 30)
                    self.pg_pool = DBPGInterface.HOST_POOLS[self.username][self.db_name]
                    try:
                        self.pg_pool.getconn(timeout=TIMEOUT)
                    except OperationalError:
                        self.close()
                        raise HttpStatusException("Database \"" + self.db_name + "\" does not exist",
                                                  CustomHTTPStatus.DATABASE_NO_EXIST)

                    conn = self._get_conn(False)
                else:
                    raise ex

        return conn

    def get_conn(self):
        try:
            conn = self._get_conn()
        except HttpStatusException as ex:
            raise ex
        except Exception:
            traceback.print_exc()
            raise HttpStatusException(ERR__connect_db, HTTPStatus.INTERNAL_SERVER_ERROR)

        return conn

    def put_conn(self, conn):
        DBPGInterface.HOST_POOLS_QUEUES[self.username][self.db_name].put((conn, self.role is not None))

    def close(self):
        DBPGInterface.HOST_POOLS[self.username].pop(self.db_name)
        DBPGInterface.HOST_POOLS_QUEUES[self.username].pop(self.db_name)
        self.pg_pool.close()

    def check_dba(self, conn, wait_hook: queue.Queue = None):
        columns, rows = self.execute_query(conn, QUERY__dba_query, parameters={KEY__database: self.db_name}, wait_hook=wait_hook)
        if not rows[0]:
            raise HttpStatusException(ERR__must_use_canned_query)

    def execute_query(self, conn, query, parameters=None, wait_hook: queue.Queue = None):
        x = 0
        err = None
        while x < PGCONN__max_conns:
            x += 1
            try:
                with conn.cursor() as cursor:
                    do_prepare = False

                    if wait_hook:
                        res, err, code = wait_hook.get()
                        if not res:
                            if code == 500:
                                raise Exception(err)
                            raise HttpStatusException(err, code)

                    if parameters is None or len(parameters.keys()) == 0:
                        cursor.execute(query, prepare=do_prepare)
                    else:
                        cursor.execute(query, parameters, prepare=do_prepare)
                    if cursor.description is None:
                        return [], []
                    else:
                        return [desc[0] for desc in cursor.description], cursor.fetchall()
            except OperationalError as ex:
                if ex.sqlstate is None or ex.sqlstate.startswith("08"):
                    self.pg_pool.putconn(conn)
                    self.pg_pool.check()
                    conn = self.get_conn()
                    err = ex
                else:
                    raise ex
            except Exception as ex:
                if self.output_query_exceptions:
                    traceback.print_exc()
                raise ex

        if err:
            raise err

    def commit(self, conn):
        conn.commit()

    def rollback(self, conn):
        conn.rollback()

    def handle_db_error(self, err, echo):
        if isinstance(err, ProgrammingError) and hasattr(err, 'pgresult'):
            err = err.pgresult.error_message.decode("ASCII")

        err = str(err)
        if echo != ECHO__none:
            err += CHAR__newline + echo
        raise HttpStatusException(err, HTTPStatus.UNPROCESSABLE_ENTITY)



KEY__id = "id"
KEY__last_totp = "last_totp"
KEY__attachment_name = "name"
KEY__application = "application"
KEY__application_description = "application_description"
KEY__configuration = "configuration"
KEY__schema = "schema"
KEY__name = "name"
KEY__is_default = "is_default"
KEY__ip_id = "ip_id"
KEY__configuration_name = "name"
KEY__configuration_description = "configuration_description"
KEY__exception = "exception"
KEY__data_validation_view = "data_validation_view"
KEY__allow_uninstall = "allow_uninstall"
KEY__uninstall_key = "uninstall_key"
KEY__install_key = "install_key"
KEY__deletion_key = "deletion_key"
KEY__render_as = "render_as"
KEY__default_password = "default_password"
KEY__url = "url"
KEY__artifact_base_uri = "artifact_base_uri"
KEY__jaaql_password = "jaaql_password"
KEY__super_db_password = "super_db_password"
KEY__public_username = "public_username"
KEY__records_total = "records_total"
KEY__records_filtered = "records_filtered"
KEY__data = "data"
KEY__db_url = "db_url"
KEY__precedence = "precedence"
KEY__roles = "roles"
KEY__pre_auth_key = "pre_auth_key"
KEY__email = "email"
KEY__sort = "sort"
KEY__invite_key = "invite_key"
KEY__invite_or_poll_key = "invite_or_poll_key"
KEY__invite_poll_key = "invite_poll_key"
KEY__invite_code = "invite_code"
KEY__invite_key_status = "invite_key_status"
KEY__reset_key = "reset_key"
KEY__reset_or_poll_key = "reset_or_poll_key"
KEY__reset_poll_key = "reset_poll_key"
KEY__reset_code = "reset_code"
KEY__reset_key_status = "reset_key_status"
KEY__parameters = "parameters"
KEY__oauth_token = "oauth_token"
KEY__template = "template"
KEY__create_file = "create_file"
KEY__subject = "subject"
KEY__recipient = "recipient"
KEY__completed = "completed"
KEY__filename = "filename"
KEY__content = "content"
KEY__attachments = "attachments"
KEY__app_relative_path = "app_relative_path"
KEY__sender = "sender"
KEY__search = "search"
KEY__application_name = "name"
KEY__db_super_user_password = "db_super_user_password"
KEY__dataset = "dataset"
KEY__dataset_name = "name"
KEY__dataset_description = "dataset_description"
KEY__size = "size"
KEY__database = "database"
KEY__read_only = "read_only"
KEY__node_name = "name"
KEY__authorization = "authorization"
KEY__argument = "argument"
KEY__is_node = "is_node"
KEY__application_new_url = "new_url"
KEY__connection = "connection"
KEY__node = "node"
KEY__application_url = "url"
KEY__key = "key"
KEY__override_send_name = "override_send_name"
KEY__ip = "ip"
KEY__status = "status"
KEY__endpoint = "endpoint"
KEY__input = "input"
KEY__duration_ms = "duration_ms"
KEY__occurred = "occurred"
KEY__jaaql_enc_symmetric_key = "jaaql_enc_symmetric_key"
KEY__enc_symmetric_key = "enc_symmetric_key"
KEY__default_enc_symmetric_key = "default_enc_symmetric_key"
KEY__role = "role"
KEY__mfa_key = "mfa_key"
KEY__mfa_enabled = "mfa_enabled"
KEY__username = "username"
KEY__password = "password"
KEY__default_database = "default_database"
KEY__new_password = "new_password"
KEY__new_password_confirm = "new_password_confirm"
KEY__page = "page"
KEY__use_mfa = "use_mfa"
KEY__show_deleted = "show_deleted"
KEY__description = "description"
KEY__the_role = "the_role"
KEY__admin_password = "admin_password"
KEY__database_name = "name"
KEY__create = "create"
KEY__drop = "drop"
KEY__force_transactional = "force_transactional"
KEY__port = "port"
KEY__old_password = "old_password"
KEY__address = "address"
KEY__public_password = "public_password"
KEY__jaaql_name = "jaaql_name"
KEY__level = "level"
KEY__code = "code"
KEY__access_rank = "access_rank"
KEY__otp_uri = "otp_uri"
KEY__otp_qr = "otp_qr"
KEY__default_otp_uri = "default_otp_uri"
KEY__default_otp_qr = "default_otp_qr"
KEY__superjaaql_otp_uri = "superjaaql_otp_uri"
KEY__superjaaql_otp_qr = "superjaaql_otp_qr"
KEY__is_public = "is_public"
KEY__count = "count"
KEY__account = "account"
KEY__enc_username = "enc_username"
KEY__encrypted_password = "encrypted_password"
KEY__email_template_name = "name"
KEY__column_name = "column_name"
KEY__table_name = "table_name"
KEY__is_primary = "is_primary"
KEY__data_validation_table = "data_validation_table"
KEY__recipient_validation_view = "recipient_validation_view"
KEY__email_template = "template"
KEY__do_send_signup_email = "do_send_signup_email"
KEY__allow_confirm_signup_attempt = "allow_confirm_signup_attempt"
KEY__allow_reset_password = "allow_reset_password"
KEY__already_signed_up_email_template = "existing_user_template"
KEY__allow_signup = "allow_signup"
KEY__email_account_name = "name"
KEY__email_sent = "sent"
KEY__password_hash = "password_hash"
KEY__user_id = "user_id"
KEY__body = "body"
KEY__default_email_signup_template = "default_email_signup_template"
KEY__default_email_already_signed_up_template = "default_email_already_signed_up_template"
KEY__default_reset_password_template = "default_reset_password_template"
KEY__document_id = "document_id"
KEY__as_attachment = "as_attachment"
KEY__email_account_protocol = "protocol"
KEY__email_account_host = "host"
KEY__email_account_port = "port"
KEY__email_account_username = "username"
KEY__email_account_send_name = "send_name"
KEY__encrypted_address = "encrypted_address"
KEY__password_id = "password_id"
KEY__created = "created"
KEY__source_file = "source_file"
KEY__position = "position"
KEY__file = "file"
KEY__error = "error"
KEY__error_row_number = "row_number"
KEY__error_index = "index"
KEY__error_query = "query"
KEY__error_set = "set"

ATTR__deleted = "deleted"

HEADER__security = "Authentication-Token"
HEADER__security_bypass = "Authentication-Token-Bypass"

SQL__where = "WHERE"
SQL__order_by = "ORDER BY"
SQL__limit = "LIMIT"
SQL__offset = "OFFSET"
SQL__asc = "ASC"
SQL__desc = "DESC"
SQL__and = "AND"
SQL__or = "OR"
SQL__like = "LIKE"
SQL__eq = "="
SQL__paren_open = "("
SQL__paren_close = ")"
SQL__single_quote = "'"
SQL__double_quote = '"'
SQL__lower = "lower"
SQL__varchar_cast = "::varchar"
SQL__is_null = "is null"

SEPARATOR__comma_space = ", "
SEPARATOR__comma = ","
SEPARATOR__space = " "

JAAQL__arg_marker = ":"

DIR__config = "config"
DIR__www = "www"
DIR__render_template = "rendered_documents"

FILE__config = "config.ini"
FILE__canned_queries = "canned_queries.jsql"

ENCODING__utf = "UTF-8"
ENCODING__ascii = "ascii"
EXAMPLE__jwt = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.e30.yXvILkvUUCBqAFlAv6wQ1Q-QRAjfe3eSosO949U73Vo"

VAULT_KEY__db_crypt_key = "db_crypt_key"
VAULT_KEY__db_repeatable_salt = "repeatable_salt"
VAULT_KEY__jaaql_local_access_key = "jaaql_access_key"
VAULT_KEY__super_db_credentials = "super_db_credentials"
VAULT_KEY__jaaql_lookup_connection = "jaaql_lookup_connection"
VAULT_KEY__allow_jaaql_uninstall = "Allow jaaql uninstall"

ENVIRON__vault_key = "JAAQL_VAULT_PASSWORD"
ENVIRON__local_install = "JAAQL_LOCAL_INSTALL"
ENVIRON__jaaql_profiling = "JAAQL_PROFILING"
ENVIRON__email_credentials = "JAAQL_EMAIL_CREDENTIALS"
ENVIRON__install_path = "INSTALL_PATH"
ENVIRON__sentinel_url = "SENTINEL_URL"
ENVIRON__canned_queries = "CANNED_QUERIES"

EMAIL_PARAM__signup_key = "JAAQL_INVITE_KEY"
EMAIL_PARAM__invite_code = "JAAQL_INVITE_SHORT_CODE"
EMAIL_PARAM__app_url = "APP_URL"

EMAIL_PARAM__reset_key = "JAAQL_RESET_KEY"
EMAIL_PARAM__reset_code = "JAAQL_RESET_SHORT_CODE"

CONFIG_KEY__server = "SERVER"
CONFIG_KEY_SERVER__port = "port"

JWT_PURPOSE__oauth = "oauth"
JWT_PURPOSE__pre_auth = "pre_auth"
JWT_PURPOSE__connection = "connection"

ERR__invalid_token = "Invalid token!"
ERR__not_yet_installed = "JAAQL has not yet been installed!"
ERR__user_public = "Cannot perform this action on a public user!"
ERR__too_many_signup_attempts = "Too many signup attempts"
ERR__too_many_reset_requests = "Too many reset requests"
ERR__too_many_code_attempts = "Code disabled due to too many incorrect attempts. Please use the link in the email"
ERR__document_still_rendering = "Document still rendering"
ERR__document_id_not_found = "Document id not found"

PG_ENV__password = "POSTGRES_PASSWORD"

HTML__base64_png = "data:image/png;base64,"
FORMAT__png = "png"

NODE__host_node = "host"
DB__jaaql = "jaaql"
DB__postgres = "postgres"

PORT__ems = 6061

ENDPOINT__configuration_authorizations = "/internal/applications/configurations/authorizations"
ENDPOINT__reload_accounts = "/reload-accounts"
ENDPOINT__send_email = "/send-email"
ENDPOINT__internal_applications = "/internal/applications"
ENDPOINT__internal_templates = "/internal/emails/templates"
ENDPOINT__internal_accounts = "/internal/emails/accounts"
ENDPOINT__is_alive = "/internal/is-alive"
ENDPOINT__jaaql_emails = "/emails"
ENDPOINT__report_sentinel_error = "/sentinel/reporting/error"
ENDPOINT__install = "/internal/install"

CONFIG__default = "Default config"
CONFIG__default_desc = "Default config description"
DATASET__default = "Default dataset"
DATASET__default_desc = "Default dataset description"

HTTP_STATUS__too_early = 425  # Shiv for python 3.8

USERNAME__jaaql = "jaaql"
USERNAME__super_db = "super_db"
ROLE__jaaql = "jaaql"
ROLE__postgres = "postgres"

PROTOCOL__postgres = "postgresql://"

VERSION = "4.3.0"




import sys

from jaaql.utilities.vault import Vault, DIR__vault
from jaaql.db.db_interface import DBInterface
import jaaql.utilities.crypt_utils as crypt_utils
from jaaql.migrations.migrations import run_migrations
from jaaql.email.email_manager import EmailManager
from jaaql.db.db_utils import execute_supplied_statement, create_interface
import hashlib
from jaaql.constants import *
from jaaql.utilities.options import OPT_KEY__canned_queries
from jaaql.config_constants import *
from os.path import dirname
from jaaql.services.cached_canned_query_service import CachedCannedQueryService
import threading

import uuid
import traceback
import subprocess
import json
import time
import os
from os.path import join

from jaaql.exceptions.http_status_exception import *
from typing import Union

CONFIG_KEY_security = "SECURITY"
CONFIG_KEY_SECURITY__mfa_label = "mfa_label"
CONFIG_KEY_SECURITY__mfa_issuer = "mfa_issuer"
CONFIG_KEY_SECURITY__do_audit = "do_audit"
CONFIG_KEY_SECURITY__token_expiry_ms = "token_expiry_ms"
CONFIG_KEY_SECURITY__token_refresh_expiry_ms = "token_refresh_expiry_ms"

DELETION_KEY__cooldown_removal = 0.1
DELETION_KEY__default_expiry_seconds = 30

ERR__expected_matcher = "'like' or '='"
ERR__expected_attribute = "attribute name"
ERR__expected_operand = "operand"
ERR__expected_logic = "'AND' or 'OR'"

ERR__expected_parser = "Expected %s in search parameter not '%s'"
ERR__deletion_invalid_key = "Deletion key invalid. Either didn't exist or expired"
ERR__deletion_key_expired = "Deletion key expired"
ERR__missing_page_size = "Cannot provide one or the other for page/size. Please provide both or neither"
ERR__unexpected_paren_close = "Cannot process ')' in search as no associated '('"
ERR__unexpected_parse_err = "Unexpected error whilst parsing search string. Please inspect"
ERR__unsupported_direction = "Sort direction should be ASC, DESC or left empty"
ERR__unsupported_column_names = "Please format sorts like 'colname ASC,othercol DESC'. Spaced columns not yet supported"
ERR__unexpected_quote = "Unexpected single quote. Only usable within the operand (after 'like' or '=')"

EXPECTING__attr = 0
EXPECTING__matcher = 1
EXPECTING__operand = 2
EXPECTING__logic = 3
EXPECTING__total = 4

WHERE__id = "where_query_"

KEY__db_crypt_key = "db_crypt_key"

VAULT_KEY__jwt_crypt_key = "jwt_crypt_key"
VAULT_KEY__jwt_obj_crypt_key = "jwt_obj_crypt_key"

FILE__was_installed = "was_installed"

JWT__data = "data"

WARNING__uninstall_allowed = "Due to installation parameters, the system can be uninstalled completely via the API. We do not recommend this being " \
                             "used in production systems (the super user db password is required to perform uninstallation so it is not 'open') "

DIR__apps = "apps"
SEPARATOR__dir = "/"


class JAAQLPivotData:
    def __init__(self, level_data: list, matching: list = None):
        self.matching = matching
        self.level_data = level_data
        self.matching = matching

        self.copy_from_idx = None
        self.copy_to_idx = None

        self.check_idxs = None


class JAAQLPivotInfo:
    def __init__(self, name, keys: [str], on: str):
        self.name = name
        self.keys = keys
        if isinstance(keys, str):
            self.keys = [keys]
        self.on = on
        if not self.on.startswith(self.name + "_"):
            self.on = self.name + "_" + self.on


class BaseJAAQLModel:

    def splice_into(self, x: [dict], y: [dict]):
        for i in range(len(x)):
            missing_dict_fields = [key for key in y[i].keys() if key not in x[i]]
            matching_dict_fields = [key for key in y[i].keys() if key in x[i]]
            for missing in missing_dict_fields:
                x[i][missing] = y[i][missing]
            for matching in matching_dict_fields:
                if isinstance(x[i][matching], list):
                    self.splice_into(x[i][matching], y[i][matching])

    def exit_jaaql(self):
        """
        Will terminate the worker forcefully which fires a hook in gunicorn_config.py
        This hook reloads all workers if the file exists in vault
        :return:
        """
        if self.is_container:
            open(join(DIR__vault, FILE__was_installed), 'a').close()
        if os.environ.get(ENVIRON__install_path, "").strip() == "/component-test-project":
            pid = open("app.pid", "r").read()
            subprocess.call("kill -HUP " + pid, shell=True)  # Kill gunicorn so coverage report is generated
        else:
            time.sleep(1)
            os._exit(0)

    def group(self, data: [dict], key_field_name: str):
        outer_dict = {}

        for itm in data:
            list_itm = [vals for _, vals in itm.items() if isinstance(vals, list)]
            val_itm = [key for key, val in itm.items() if key != key_field_name]
            if len(list_itm) == 0:
                if len(val_itm) == 1:
                    outer_dict[itm[key_field_name]] = itm[val_itm[0]]
                else:
                    outer_dict[itm[key_field_name]] = {key: itm[key] for key in val_itm}
            else:
                outer_dict[itm[key_field_name]] = self.group(list_itm[0], key_field_name)

        return outer_dict

    def pivot(self, data: [dict], pivot_info: [JAAQLPivotInfo]):
        if len(data) == 0:
            return data
        if not isinstance(pivot_info, list):
            pivot_info = [pivot_info]

        ret = []
        stack = [JAAQLPivotData(ret)]
        cols = list(data[0].keys())

        for row in data:
            copy_from_idx = 0
            last_pivot_length = -1
            for pivot, pivot_idx in zip(pivot_info, range(len(pivot_info))):
                copy_to_idx = cols.index(pivot.on)

                level_data = {key[last_pivot_length + 1:]: val for key, val in row.items() if copy_from_idx <=
                              cols.index(key) < copy_to_idx}
                matching = [level_data[key[last_pivot_length + 1:]] for key in pivot.keys]
                is_empty = len([match for match in matching if match is not None]) == 0

                cur = stack[pivot_idx]
                sub_list = cur.level_data
                if cur.matching != matching:
                    if not is_empty:
                        sub_list.append(level_data)
                    cur.matching = matching
                    stack = stack[0:pivot_idx + 1]

                    new_data = []
                    level_data[pivot.name] = new_data
                    stack.append(JAAQLPivotData(new_data))

                copy_from_idx = copy_to_idx
                last_pivot_length = len(pivot.name)

            cur = stack[-1]
            final_data = {key[last_pivot_length + 1:]: val for key, val in row.items() if copy_from_idx <=
                          cols.index(key)}
            is_empty = len([key for key in final_data if final_data[key] is not None]) == 0
            if not is_empty:
                cur.level_data.append(final_data)

        return ret

    def set_jaaql_lookup_connection(self):
        if self.vault.has_obj(VAULT_KEY__jaaql_lookup_connection):
            jaaql_uri = self.vault.get_obj(VAULT_KEY__jaaql_lookup_connection)
            address, port, db, username, password = DBInterface.fracture_uri(jaaql_uri)
            self.jaaql_lookup_connection = create_interface(self.config, address, port, db, username, password=password)
            if self.options.get(OPT_KEY__canned_queries) or os.environ.get(ENVIRON__canned_queries) == "TRUE":
                self.cached_canned_query_service = CachedCannedQueryService(self.is_container, self.jaaql_lookup_connection)

    def __init__(self, config, vault_key: str, options: dict, migration_project_name: str = None,
                 migration_folder: str = None, is_container: bool = False, url: str = None):
        self.config = config
        self.migration_project_name = migration_project_name
        self.migration_folder = migration_folder
        self.is_container = is_container

        self.cached_canned_query_service = None

        self.url = url
        self.has_installed = False

        self.force_mfa = config[CONFIG_KEY__security][CONFIG_KEY_SECURITY__force_mfa]
        self.do_audit = config[CONFIG_KEY__security][CONFIG_KEY_SECURITY__do_audit]

        self.vault = Vault(vault_key, DIR__vault)
        self.jaaql_lookup_connection = None
        self.email_manager = None

        self.options = options

        self.token_expiry_ms = int(config[CONFIG_KEY__security][CONFIG_KEY_SECURITY__token_expiry_ms])
        self.refresh_expiry_ms = int(config[CONFIG_KEY__security][CONFIG_KEY_SECURITY__token_refresh_expiry_ms])

        if not self.vault.has_obj(VAULT_KEY__db_repeatable_salt):
            self.vault.insert_obj(VAULT_KEY__db_repeatable_salt, crypt_utils.fetch_random_readable_salt().decode(crypt_utils.ENCODING__ascii))

        if not self.vault.has_obj(VAULT_KEY__db_crypt_key):
            _, db_crypt_key = crypt_utils.key_stretcher(str(uuid.uuid4()), length=crypt_utils.AES__key_length)
            self.vault.insert_obj(VAULT_KEY__db_crypt_key, db_crypt_key.decode(crypt_utils.ENCODING__ascii))

        if not self.vault.has_obj(VAULT_KEY__jwt_crypt_key):
            _, jwt_crypt_key = crypt_utils.key_stretcher(str(uuid.uuid4()))
            self.vault.insert_obj(VAULT_KEY__jwt_crypt_key, jwt_crypt_key.decode(crypt_utils.ENCODING__ascii))

        if not self.vault.has_obj(VAULT_KEY__jwt_obj_crypt_key):
            _, jwt_obj_crypt_key = crypt_utils.key_stretcher(str(uuid.uuid4()))
            self.vault.insert_obj(VAULT_KEY__jwt_obj_crypt_key, jwt_obj_crypt_key.decode(crypt_utils.ENCODING__ascii))

        if self.vault.has_obj(VAULT_KEY__jaaql_lookup_connection):
            if self.vault.has_obj(VAULT_KEY__allow_jaaql_uninstall):
                self.uninstall_key = str(uuid.uuid4())
                print(WARNING__uninstall_allowed, file=sys.stderr)
                print("UNINSTALL KEY: " + self.uninstall_key)
            else:
                self.install_key = None
            self.has_installed = True

        self.set_jaaql_lookup_connection()

        self.reload_lock = threading.Lock()
        self.symmetric_keys = None

        if self.vault.has_obj(VAULT_KEY__jaaql_lookup_connection):
            run_migrations(self.jaaql_lookup_connection, self.is_container, options=options, key=self.get_db_crypt_key())

            if migration_project_name is not None:
                run_migrations(self.jaaql_lookup_connection, self.is_container, migration_project_name, migration_folder=migration_folder,
                               config=self.config, super_credentials=self.vault.get_obj(VAULT_KEY__super_db_credentials), options=options,
                               key=self.get_db_crypt_key())

            if self.is_container:
                self.jaaql_lookup_connection.close()  # Each individual class will have one

            self.email_manager = EmailManager(self.is_container)
        else:
            self.install_key = str(uuid.uuid4())
            with open(os.path.join(dirname(dirname(dirname(__file__))), "install_key"), "w") as install_key_file:
                install_key_file.write(self.install_key)
            print("INSTALL KEY: " + self.install_key)

        if not self.vault.has_obj(VAULT_KEY__jaaql_local_access_key):
            self.vault.insert_obj(VAULT_KEY__jaaql_local_access_key, str(uuid.uuid4()))

        self.local_access_key = self.vault.get_obj(VAULT_KEY__jaaql_local_access_key)

    def get_db_crypt_key(self):
        return self.vault.get_obj(VAULT_KEY__db_crypt_key).encode(crypt_utils.ENCODING__ascii)

    def get_repeatable_salt(self, addition: str = None):
        repeatable = self.vault.get_obj(VAULT_KEY__db_repeatable_salt)
        if addition is not None:
            return hashlib.md5((repeatable + addition).encode(crypt_utils.ENCODING__ascii)).digest()
        else:
            return repeatable.encode(crypt_utils.ENCODING__ascii)

    def setup_paging_parameters(self, inputs: dict, has_deleted: bool = True):
        inputs_no_del = inputs.copy()
        if KEY__show_deleted in inputs_no_del:
            inputs_no_del.pop(KEY__show_deleted)
        paging_dict, parameters = self.format_paging_request(inputs_no_del)

        if has_deleted:
            show_deleted = inputs.get(KEY__show_deleted, False)
            deleted_condition = ATTR__deleted + SEPARATOR__space + SQL__is_null
            if paging_dict[SQL__where] is None and not show_deleted:
                paging_dict[SQL__where] = deleted_condition
            elif paging_dict[SQL__where] is not None and not show_deleted:
                existing = SQL__paren_open + paging_dict[SQL__where] + SQL__paren_close
                paging_dict[SQL__where] = existing + SEPARATOR__space + SQL__and + SEPARATOR__space + deleted_condition

        return paging_dict, parameters

    def get_default_app_url(self):
        return self.url + SEPARATOR__dir + DIR__apps

    def replace_default_app_url(self, url_with_default: str):
        if url_with_default is None:
            return None
        return url_with_default.replace("{{DEFAULT}}", self.get_default_app_url())

    def execute_paging_query(self, jaaql_connection: DBInterface, full_query: str, count_query: str, parameters: dict,
                             where_query: str, where_parameters: dict, decrypt_columns: list = None,
                             encryption_key: bytes = None):
        data = execute_supplied_statement(jaaql_connection, full_query, parameters=parameters,
                                          as_objects=True, decrypt_columns=decrypt_columns,
                                          encryption_key=encryption_key)
        total = execute_supplied_statement(jaaql_connection, count_query, as_objects=True)
        total = total[0]["count"]
        total_filtered = execute_supplied_statement(jaaql_connection, count_query + where_query,
                                                    parameters=where_parameters,
                                                    as_objects=True)[0]["count"]
        return self.paged_collection(total, total_filtered, data)

    @staticmethod
    def paged_collection(total: int, filtered: int, data: list):
        return {
            KEY__records_total: total,
            KEY__records_filtered: filtered,
            KEY__data: data
        }

    @staticmethod
    def construct_paging_queries(http_inputs: dict) -> (str, str, dict, dict):
        formatted, parameters = BaseJAAQLModel.format_paging_request(http_inputs)
        qa, qaw, wp = BaseJAAQLModel.construct_formatted_paging_queries(formatted, parameters)
        return qa, qaw, wp, parameters

    @staticmethod
    def construct_formatted_paging_queries(paging_input: dict, paging_parameters: dict) -> (str, str, dict):
        query_addition = ""
        query_addition_where = ""

        if paging_input[SQL__where] is not None:
            query_addition += SEPARATOR__space + SQL__where + SEPARATOR__space + paging_input[SQL__where]
            query_addition_where = query_addition
        if paging_input[SQL__order_by] is not None:
            query_addition += SEPARATOR__space + SQL__order_by + SEPARATOR__space + paging_input[SQL__order_by]
        if paging_input[SQL__limit] is not None:
            query_addition += SEPARATOR__space + SQL__limit + SEPARATOR__space + str(paging_input[SQL__limit])
        if paging_input[SQL__offset] is not None:
            query_addition += SEPARATOR__space + SQL__offset + SEPARATOR__space + str(paging_input[SQL__offset])

        where_params = {}
        for param in paging_parameters.keys():
            if param.startswith(WHERE__id):
                where_params[param] = paging_parameters[param]

        return query_addition, query_addition_where, where_params

    @staticmethod
    def format_paging_request(http_inputs: dict) -> (dict, dict):
        sort_input = http_inputs.get(KEY__sort, "")
        if sort_input is None:
            sort_input = ""
        sort_broken = sort_input.replace(SEPARATOR__comma_space, SEPARATOR__comma).split(SEPARATOR__comma)
        all_parameters = {}
        order_query = []

        for cur_sort in sort_broken:
            cur_broken = cur_sort.split(SEPARATOR__space)
            if len(cur_broken) == 1:
                direction = SQL__asc
            elif len(cur_broken) > 2:
                raise HttpStatusException(ERR__unsupported_column_names)
            else:
                direction = cur_broken[1]

            direction = direction.upper()
            if direction not in [SQL__asc, SQL__desc]:
                raise HttpStatusException(ERR__unsupported_direction)

            column = cur_broken[0]
            order_query.append(SQL__double_quote + column + SQL__double_quote + SEPARATOR__space + direction)

        built_order = SEPARATOR__comma_space.join(order_query)
        if len(sort_input) == 0:
            order_query = []

        if KEY__size not in http_inputs or http_inputs[KEY__size] is None:
            size = None
        else:
            # Int not strictly needed as type checking is already done at the documentation level. This is merely
            # defensive as this method could be abused by an extension
            size = int(http_inputs[KEY__size])

        if KEY__page not in http_inputs or http_inputs[KEY__page] is None:
            page = None
        else:
            page = int(http_inputs[KEY__page])

        if (page is None and size is not None) or (page is not None and size is None):
            raise HttpStatusException(ERR__missing_page_size)

        offset = None
        if page is not None:
            offset = page * size

        try:
            where_str = BaseJAAQLModel.parse_where(http_inputs.get(KEY__search, ""), all_parameters)
        except HttpStatusException as err:
            raise err
        except Exception as err:
            traceback.print_tb(err.__traceback__)
            raise HttpStatusException(ERR__unexpected_parse_err)

        if KEY__db_crypt_key in http_inputs:
            all_parameters[KEY__db_crypt_key] = http_inputs[KEY__db_crypt_key]

        return {
                   SQL__where: where_str,
                   SQL__order_by: None if len(order_query) == 0 else built_order,
                   SQL__limit: size,
                   SQL__offset: offset
               }, all_parameters

    @staticmethod
    def check_parse_state(cur_token: str, expected_state: Union[int, list], actual_state: int):
        if isinstance(expected_state, int):
            expected_state = [expected_state]

        if actual_state in expected_state:
            return

        expected_err = []
        for expected in expected_state:
            if expected == EXPECTING__logic:
                expected_err.append(ERR__expected_logic)
            elif expected == EXPECTING__operand:
                expected_err.append(ERR__expected_operand)
            elif expected == EXPECTING__attr:
                expected_err.append(ERR__expected_attribute)
            else:
                expected_err.append(ERR__expected_matcher)

        spaced_or = SEPARATOR__space + SQL__or + SEPARATOR__space

        raise HttpStatusException(ERR__expected_parser % (spaced_or.join(expected_err), cur_token))

    @staticmethod
    def parse_where_token(parse_parameters: dict, parameters: dict):
        and_separated = SEPARATOR__space + SQL__and + SEPARATOR__space
        or_separated = SEPARATOR__space + SQL__or + SEPARATOR__space

        if parse_parameters["cur_token"].upper() == SQL__and:
            BaseJAAQLModel.check_parse_state(parse_parameters["cur_token"], EXPECTING__logic, parse_parameters["state"])
            parse_parameters["cur_search"] += and_separated
        elif parse_parameters["cur_token"].upper() == SQL__or:
            BaseJAAQLModel.check_parse_state(parse_parameters["cur_token"], EXPECTING__logic, parse_parameters["state"])
            parse_parameters["cur_search"] += or_separated
        elif parse_parameters["cur_token"] == SQL__paren_open:
            BaseJAAQLModel.check_parse_state(parse_parameters["cur_token"], EXPECTING__attr, parse_parameters["state"])
            parse_parameters["paren_depth"] += 1
            parse_parameters["cur_search"] += "("
        elif parse_parameters["cur_token"] == SQL__paren_close:
            BaseJAAQLModel.check_parse_state(parse_parameters["cur_token"], EXPECTING__logic, parse_parameters["state"])
            if parse_parameters["paren_depth"] == 0:
                raise HttpStatusException(ERR__unexpected_paren_close)
            parse_parameters["paren_depth"] -= 1
            parse_parameters["cur_search"] += ")"
        elif parse_parameters["cur_token"].upper() == SQL__like:
            BaseJAAQLModel.check_parse_state(parse_parameters["cur_token"], EXPECTING__matcher,
                                             parse_parameters["state"])
            parse_parameters["cur_search"] += SEPARATOR__space + SQL__like + SEPARATOR__space
        elif parse_parameters["cur_token"] == SQL__eq:
            BaseJAAQLModel.check_parse_state(parse_parameters["cur_token"], EXPECTING__matcher,
                                             parse_parameters["state"])
            parse_parameters["cur_search"] += SEPARATOR__space + SQL__eq + SEPARATOR__space
        else:
            BaseJAAQLModel.check_parse_state(parse_parameters["cur_token"], [EXPECTING__attr, EXPECTING__operand],
                                             parse_parameters["state"])

            if parse_parameters["state"] == EXPECTING__operand:
                where_id = WHERE__id + str(parse_parameters["param_num"])
                parameters[where_id] = parse_parameters["cur_token"]

                arg_marker = JAAQL__arg_marker + where_id
                as_attr_operand = SQL__lower + SQL__paren_open + arg_marker + SQL__paren_close + SQL__varchar_cast
                parse_parameters["param_num"] += 1
                parse_parameters["cur_search"] += as_attr_operand
            else:
                # TODO potential SQL injection issue. Is the user able to pass spaces in that don't look like spaces
                # TODO can we bind column names
                quoted = SQL__double_quote + parse_parameters["cur_token"] + SQL__double_quote + SQL__varchar_cast
                parse_parameters["cur_search"] += SQL__lower + SQL__paren_open + quoted + SQL__paren_close

        parse_parameters["state"] += 1
        parse_parameters["state"] %= EXPECTING__total

        parse_parameters["cur_token"] = ""

    @staticmethod
    def parse_where(search_str: str, parameters: dict):
        parse_parameters = {
            "cur_search": "",
            "cur_token": "",
            "state": EXPECTING__attr,
            "paren_depth": 0,
            "quote_char": None,
            "param_num": 0
        }

        if search_str is None:
            return None

        for i in range(len(search_str)):
            cur_char = search_str[i]
            if cur_char == SEPARATOR__space and parse_parameters["quote_char"] is None:
                BaseJAAQLModel.parse_where_token(parse_parameters, parameters)
            elif cur_char in [SQL__eq, SQL__paren_open, SQL__paren_close] and parse_parameters["quote_char"] is None:
                BaseJAAQLModel.parse_where_token(parse_parameters, parameters)
                parse_parameters["cur_token"] = cur_char
                BaseJAAQLModel.parse_where_token(parse_parameters, parameters)
            elif cur_char == SQL__single_quote and parse_parameters["quote_char"] is None:
                if parse_parameters["state"] != EXPECTING__operand or len(parse_parameters["cur_token"]) != 0:
                    raise HttpStatusException(ERR__unexpected_quote)
                parse_parameters["quote_char"] = SQL__single_quote
            elif cur_char == SQL__single_quote and parse_parameters["quote_char"] is SQL__single_quote:
                parse_parameters["quote_char"] = None
            else:
                parse_parameters["cur_token"] += cur_char

        if len(parse_parameters["cur_token"]) != 0:
            BaseJAAQLModel.parse_where_token(parse_parameters, parameters)

        return parse_parameters["cur_search"] if len(parse_parameters["cur_search"]) != 0 else None

    def request_deletion_key(self, purpose: str, input_args: dict,
                             expiry_seconds: int = DELETION_KEY__default_expiry_seconds):
        jwt_obj_key = self.vault.get_obj(VAULT_KEY__jwt_obj_crypt_key)
        jwt_key = self.vault.get_obj(VAULT_KEY__jwt_crypt_key)

        decoded = {
            JWT__data: crypt_utils.encrypt(jwt_obj_key, json.dumps(input_args))
        }

        return crypt_utils.jwt_encode(jwt_key, decoded, purpose, expiry_ms=expiry_seconds * 1000)

    def validate_deletion_key(self, key: str, purpose: str) -> dict:
        jwt_obj_key = self.vault.get_obj(VAULT_KEY__jwt_obj_crypt_key)
        jwt_key = self.vault.get_obj(VAULT_KEY__jwt_crypt_key)

        key = crypt_utils.jwt_decode(jwt_key, key, purpose)
        if not key:
            raise HttpStatusException(ERR__deletion_invalid_key)

        data = json.loads(crypt_utils.decrypt(jwt_obj_key, key[JWT__data]))
        return data



import threading
import traceback
import uuid
from functools import wraps
from werkzeug.exceptions import HTTPException, InternalServerError
import inspect
import json
import requests
from datetime import datetime
from jaaql.exceptions.custom_http_status import CustomHTTPStatus
import sys
import os
from queue import Queue
from jaaql.utilities.utils import time_delta_ms, Profiler
from flask import Response, Flask, request, jsonify, current_app
from jaaql.constants import *
from jaaql.mvc.model import JAAQLModel
from jaaql.mvc.response import JAAQLResponse
from typing import Union

from jaaql.openapi.swagger_documentation import SwaggerDocumentation, SwaggerMethod, TYPE__response, \
    SwaggerFlatResponse, REST__DELETE, REST__GET, REST__OPTIONS, REST__POST, REST__PUT, SwaggerList, SwaggerResponse, \
    MOCK__description, ARG_RESP__allow_all, RES__allow_all, SwaggerArgumentResponse, SwaggerSimpleList
from jaaql.exceptions.http_status_exception import *

ARG__http_inputs = "http_inputs"
ARG__user_id = "user_id"
ARG__ip_address = "ip_address"
ARG__response = "response"
ARG__username = "username"
ARG__profiler = "profiler"
ARG__auth_token = "auth_token"
ARG__connection = "connection"
ARG__is_public = "is_public"
ARG__verification_hook = "verification_hook"
ARG_START__connection = "connection__"
ARG_START__jaaql_connection = "jaaql_" + ARG_START__connection

CONTENT__encoding = "charset=utf-8"
CONTENT__json = "application/json"

CORS__WILDCARD = "*"

ERR__argument_wrong_type = "Argument '%s' is of type '%s' but should be of type '%s'"
ERR__duplicated_field = "Field duplicated in request"
ERR__expected_argument = "Expected argument '%s'"
ERR__expected_json = "Expected content-type " + CONTENT__json
ERR__expected_response_dict = "Expected response to be a dictionary type"
ERR__expected_response_list = "Expected response to be a array type"
ERR__expected_response_flat_type = "Expected response to be a flat type"
ERR__expected_response_field = "Expected response field '%s'"
ERR__expected_utf8 = "Please use UTF-8 encoding"
ERR__req_method_resp = "Please set the method response in the documentation if the model returns nothing"
ERR__response_wrong_type = "Response field '%s' is of type '%s' but should be of type '%s'"
ERR__unexpected_argument = "Unexpected argument '%s'"
ERR__unexpected_request_body = "Unexpected request body"
ERR__unexpected_response_code = "Response with code '%d' was not expected"
ERR__unexpected_response_field = "Unexpected response field '%s'"
ERR__method_required_security = "Method requires connection input yet marked as not secure in documentation"
ERR__method_required_token = "Method requires oauth token input yet marked as not secure in documentation"
ERR__method_required_connection = "Method requires connection input yet marked as not secure in documentation"
ERR__method_required_is_public = "Method requires is_public input yet marked as not secure in documentation"
ERR__method_required_user_id = "Method requires user id input yet marked as not secure in documentation"
ERR__method_required_user_connection = "Method requires user connection input yet marked as not secure in documentation"
ERR__method_required_username = "Method requires username yet marked as not secure in documentation"
ERR__sentinel_failed = "Sentinel failed. Reponse code '%d' and content '%s'"

FLASK__json_sort_keys = "JSON_SORT_KEYS"
FLASK__max_content_length = "MAX_CONTENT_LENGTH"

HEADER__allow_headers = "Access-Control-Allow-Headers"
HEADER__allow_origin = "Access-Control-Allow-Origin"
HEADER__allow_methods = "Access-Control-Allow-Methods"
HEADER__real_ip = "X-Real-IP"

BOOL__allowed = {
    "True": True,
    "False": False,
    "true": True,
    "false": False
}


class BaseJAAQLController:

    sentinel_errors = None
    internal_sentinel = False

    def __init__(self, model: JAAQLModel, is_prod: bool, base_url: str, do_profiling: bool = False):
        super().__init__()
        self.app = Flask(__name__, instance_relative_config=True)
        self.app.config[FLASK__json_sort_keys] = False
        self.app.config[FLASK__max_content_length] = 1024 * 1024 * 2  # 2 MB
        self._init_error_handlers(self.app)
        self.model = model
        self.do_profiling = do_profiling
        self.documentation_memory = {}
        self.profiling_request_ids = {}
        self.app.model = model
        self.is_prod = is_prod
        BaseJAAQLController.sentinel_errors = Queue()
        self.sentinel_url = os.environ.get(ENVIRON__sentinel_url)
        if self.sentinel_url:
            if self.sentinel_url == "_":
                self.sentinel_url = base_url + ENDPOINT__report_sentinel_error
                BaseJAAQLController.internal_sentinel = True
            else:
                if not self.sentinel_url.startswith("http"):
                    self.sentinel_url = "https://" + self.sentinel_url
                if not self.sentinel_url.endswith("/api") and not self.sentinel_url.endswith(ENDPOINT__report_sentinel_error):
                    self.sentinel_url = self.sentinel_url + "/api"
                if not self.sentinel_url.endswith(ENDPOINT__report_sentinel_error):
                    self.sentinel_url = self.sentinel_url + ENDPOINT__report_sentinel_error

            threading.Thread(target=self.sentinel_reporter).start()

    def sentinel_reporter(self):
        while True:
            try:
                se = BaseJAAQLController.sentinel_errors.get()
                res = requests.post(self.sentinel_url, json=se)
                if res.status_code != HTTPStatus.OK:
                    raise Exception(ERR__sentinel_failed % (res.status_code, res.text))
            except:
                traceback.print_exc()

    def diff_ms(self, start, now):
        return round((now - start).total_seconds() * 1000)

    @staticmethod
    def enforce_content_type_json():
        if request.content_type is None or request.content_type.split(";")[0] != CONTENT__json:
            raise HttpStatusException(ERR__expected_json, HTTPStatus.BAD_REQUEST)
        if len(request.content_type.split(";")) > 1:
            if request.content_type.split(";")[1].strip().lower() != CONTENT__encoding:
                raise HttpStatusException(ERR__expected_utf8, HTTPStatus.BAD_REQUEST)

    @staticmethod
    def validate_data_rec(arguments: [SwaggerArgumentResponse], data: dict, is_prod: bool, fill_missing: bool = True):
        for arg in arguments:
            if arg.required is True and arg.name not in data:
                if arg.local_only and is_prod:
                    if fill_missing:
                        data[arg.name] = None
                    continue
                raise HttpStatusException(ERR__expected_argument % arg.name, HTTPStatus.BAD_REQUEST)

            if isinstance(arg.arg_type, SwaggerList) and arg.name in data:
                if not isinstance(data[arg.name], list):
                    raise HttpStatusException(ERR__argument_wrong_type % (arg.name, str(type(data[arg.name])),
                                                                          str(type(list))), HTTPStatus.BAD_REQUEST)
                for itm in data[arg.name]:
                    BaseJAAQLController.validate_data_rec(arg.arg_type.responses, itm, fill_missing)
            elif arg.name in data and arg.arg_type == ARG_RESP__allow_all:
                if not isinstance(data[arg.name], dict):
                    raise HttpStatusException(ERR__argument_wrong_type % (arg.name, str(type(data[arg.name])),
                                                                          str(type(dict))), HTTPStatus.BAD_REQUEST)
            elif arg.name in data and not isinstance(data[arg.name], arg.arg_type):
                if data[arg.name] is None and not arg.required:
                    continue
                was_err = True
                if arg.arg_type == int:
                    try:
                        test = int(data[arg.name])
                        if str(test) == data[arg.name]:
                            was_err = False
                            data[arg.name] = test
                    except ValueError:
                        pass
                elif arg.arg_type == float:
                    try:
                        test = float(data[arg.name])
                        if str(test) == data[arg.name]:
                            was_err = False
                            data[arg.name] = test
                    except ValueError:
                        pass
                elif arg.arg_type == bool:
                    try:
                        if data[arg.name] in BOOL__allowed:
                            was_err = False
                            data[arg.name] = BOOL__allowed[data[arg.name]]
                    except ValueError:
                        pass

                if was_err:
                    raise HttpStatusException(ERR__argument_wrong_type % (arg.name, str(type(data[arg.name])),
                                                                          str(arg.arg_type)), HTTPStatus.BAD_REQUEST)
            elif arg.name not in data and fill_missing:
                data[arg.name] = None
            elif arg.arg_type == str and arg.name in data and data[arg.name] is not None:
                if arg.strip:
                    data[arg.name] = data[arg.name].strip()
                if arg.lower:
                    data[arg.name] = data[arg.name].lower()

        for key, _ in data.items():
            found = any([arg.name == key for arg in arguments])
            if not found:
                raise HttpStatusException(ERR__unexpected_argument % key, HTTPStatus.BAD_REQUEST)

    @staticmethod
    def validate_data(method: SwaggerMethod, data: dict, is_prod: bool = False, fill_missing: bool = True):
        BaseJAAQLController.validate_data_rec(method.arguments + method.body, data, is_prod, fill_missing)

    @staticmethod
    def get_method(swagger_documentation: SwaggerDocumentation):
        for method in swagger_documentation.methods:
            if request.method == method.method:
                return method

    @staticmethod
    def get_response(method: SwaggerMethod, status: Union[HTTPStatus, Namespace, int]) -> TYPE__response:
        match_value = status
        if not isinstance(status, int):
            match_value = status.value
        for resp in method.responses:
            if isinstance(resp.code, int):
                if resp.code == match_value:
                    return resp
            else:
                if resp.code.value == match_value:
                    return resp

        raise Exception(ERR__unexpected_response_code % status.value)

    @staticmethod
    def bi_cast(real_resp, name, arg_type: type):
        """
        Attempt to cast a response to it's intended type. Then check equality when casting back. If this can happen e.g.
        '5' can be cast to 5 and then back to '5' it's considered equal and the cast is performed
        :return:
        """
        err_mess = ERR__response_wrong_type % (str(name), str(type(real_resp[name])), str(arg_type))
        bi_cast = None
        try:
            # Bi directional cast. Cast to expected type and then cast back
            bi_cast = type(real_resp[name])(arg_type(real_resp[name]))
        except:
            pass
        if real_resp[name] == bi_cast:
            real_resp[name] = arg_type(real_resp[name])
        else:
            raise Exception(err_mess)

    @staticmethod
    def validate_output(response: TYPE__response, real_resp: any) -> any:
        if isinstance(response, SwaggerFlatResponse):
            check_resp = response.body
            if isinstance(real_resp, dict) or isinstance(real_resp, list):
                raise Exception(ERR__expected_response_flat_type)

            if real_resp is None and check_resp is not None:
                real_resp = check_resp
            elif real_resp is None and check_resp is None:
                raise Exception(ERR__req_method_resp)
            elif real_resp is not None and check_resp is None:
                pass  # do not update resp with method response body
            else:
                pass  # take the supplied response as gospel
        elif isinstance(response.responses, SwaggerList):
            if not isinstance(real_resp, list):
                raise Exception(ERR__expected_response_list)
            else:
                for idx in range(len(real_resp)):
                    mock_resp = SwaggerResponse(MOCK__description, response=response.responses.responses)
                    real_resp[idx] = BaseJAAQLController.validate_output(mock_resp, real_resp[idx])
        elif isinstance(response.responses, SwaggerSimpleList):
            if not isinstance(real_resp, list):
                raise Exception(ERR__expected_response_list)
            else:
                for idx in range(len(real_resp)):
                    if isinstance(real_resp[idx], datetime):
                        real_resp[idx] = str(real_resp[idx])
                    BaseJAAQLController.bi_cast(real_resp, idx, response.responses.arg_type)
        else:
            if not isinstance(real_resp, dict):
                raise Exception(ERR__expected_response_dict)

            if response.error_on_unexpected_field:
                for key, _ in real_resp.items():
                    if not any([swag_resp.name == key for swag_resp in response.responses]):
                        raise Exception(ERR__unexpected_response_field % key)

            match_alpha_resp = {}
            for swag_resp in response.responses:
                is_complex = isinstance(swag_resp.arg_type, SwaggerList)
                if swag_resp.name not in real_resp:
                    if swag_resp.required:
                        raise Exception(ERR__expected_response_field % swag_resp.name)
                    else:
                        continue
                if isinstance(real_resp[swag_resp.name], datetime):
                    real_resp[swag_resp.name] = str(real_resp[swag_resp.name])
                if swag_resp.arg_type != ARG_RESP__allow_all:
                    if not is_complex and not isinstance(real_resp[swag_resp.name], swag_resp.arg_type):
                        if real_resp[swag_resp.name] is not None or swag_resp.required:
                            BaseJAAQLController.bi_cast(real_resp, swag_resp.name, swag_resp.arg_type)

                    if is_complex:
                        mock_resp = SwaggerResponse(MOCK__description, response=swag_resp.arg_type)
                        real_resp[swag_resp.name] = BaseJAAQLController.validate_output(mock_resp,
                                                                                        real_resp[swag_resp.name])

                match_alpha_resp[swag_resp.name] = real_resp[swag_resp.name]
            real_resp = match_alpha_resp

        return real_resp

    @staticmethod
    def get_input_as_dictionary(method: SwaggerMethod, is_prod: bool, fill_missing: bool = True):
        data = {}

        was_allow_all = False
        if len(method.arguments) != 0:
            if method.arguments[0] == ARG_RESP__allow_all:
                was_allow_all = True

        if len(method.body) != 0 or was_allow_all:
            BaseJAAQLController.enforce_content_type_json()
            data = request.json
        else:
            if len(request.data.decode(request.charset)) != 0:
                raise HttpStatusException(ERR__unexpected_request_body, HTTPStatus.BAD_REQUEST)

        if isinstance(data, list):
            combined_data = data
        else:
            combined_data = {**request.form, **request.args, **data}

            if len(combined_data) != len(request.form) + len(request.args) + len(data):
                raise HttpStatusException(ERR__duplicated_field, HTTPStatus.BAD_REQUEST)

        if not was_allow_all:
            BaseJAAQLController.validate_data(method, combined_data, is_prod, fill_missing)

        return combined_data

    @staticmethod
    def _cors(resp):
        resp.headers.add(HEADER__allow_origin, CORS__WILDCARD)
        resp.headers.add(HEADER__allow_headers, CORS__WILDCARD)
        resp.headers.add(HEADER__allow_methods, CORS__WILDCARD)
        return resp

    def log_safe_dump_recursive(self, data):
        if isinstance(data, list):
            return [self.log_safe_dump_recursive(itm) for itm in data]
        elif isinstance(data, dict):
            return {
                idx: "*******" if "password" in idx.lower() else itm
                for idx, itm in data.items()
            }
        else:
            return data

    def perform_profile(self, profile_id, description: str = None, route: str = None, method: str = None):
        if not self.do_profiling:
            return

        if profile_id in self.profiling_request_ids:
            if description is None:
                raise Exception("Expected profiling description")
            if route is not None:
                raise Exception("Expected no profiling route")
            if method is not None:
                raise Exception("Expected no profiling method")
            cur_time = str(time_delta_ms(self.profiling_request_ids[profile_id], datetime.now()))
            print("PROFILING: " + str(profile_id) + " " + cur_time + " " + description)
            self.profiling_request_ids[profile_id] = datetime.now()
        elif description is not None:
            raise Exception("Expected no profiling description")
        else:
            if route is None:
                raise Exception("Expected profiling route")
            if method is None:
                raise Exception("Expected profiling method")
            print("PROFILING: " + str(profile_id) + " " + method + " " + route)
            self.profiling_request_ids[profile_id] = datetime.now()

    def log_safe_dump(self, data):
        """
        Performs a log safe json dump of the data
        :return:
        """
        return json.dumps(self.log_safe_dump_recursive(data))

    def cors_route(self, route: str, swagger_documentation: Union[list, SwaggerDocumentation]):
        documentation_as_lists = swagger_documentation
        if not isinstance(documentation_as_lists, list):
            documentation_as_lists = [documentation_as_lists]

        methods = []
        for cur_documentation in documentation_as_lists:
            cur_documentation.path = route
            for method in cur_documentation.methods:
                methods.append(method.method)

        methods.append(REST__OPTIONS)
        swagger_documentation = documentation_as_lists[0]

        def wrap_func(view_func):
            @wraps(view_func)
            def routed_function(view_func_local):
                if not self.model.has_installed and not route.startswith('/internal'):
                    resp = Response("Still installing. Be patient", status=HTTPStatus.SERVICE_UNAVAILABLE)
                    self._cors(resp)
                    return resp
                start_time = datetime.now()
                request_id = uuid.uuid4()
                self.perform_profile(request_id, route=route, method=request.method)
                resp = None
                resp_type = current_app.config["JSONIFY_MIMETYPE"]
                jaaql_resp = JAAQLResponse()
                jaaql_resp.response_type = resp_type

                if not BaseJAAQLController.is_options():
                    the_method = BaseJAAQLController.get_method(swagger_documentation)
                    self.perform_profile(request_id, "Fetch method")

                    user_id = None
                    ip_id = None
                    username = None
                    is_public = None
                    verification_hook = None
                    if method.parallel_verification:
                        verification_hook = Queue()

                    ip_addr = request.headers.get(HEADER__real_ip, request.remote_addr).split(",")[0]

                    if swagger_documentation.security:
                        if verification_hook:
                            user_id, username, ip_id, is_public = self.model.verify_auth_token_threaded(request.headers.get(HEADER__security),
                                                                                                                ip_addr, verification_hook)
                            self.perform_profile(request_id, "Verify JWT Threaded")
                        else:
                            user_id, username, ip_id, is_public = self.model.verify_auth_token(request.headers.get(HEADER__security), ip_addr)
                            self.perform_profile(request_id, "Verify JWT")

                    supply_dict = {}

                    throw_ex = None
                    ex_msg = None
                    method_input = None
                    try:
                        if ARG__http_inputs in inspect.getfullargspec(view_func_local).args:
                            supply_dict[ARG__http_inputs] = BaseJAAQLController.get_input_as_dictionary(the_method, self.is_prod)
                            method_input = self.log_safe_dump(supply_dict[ARG__http_inputs])

                        if ARG__user_id in inspect.getfullargspec(view_func_local).args:
                            if not swagger_documentation.security:
                                raise Exception(ERR__method_required_user_id)
                            supply_dict[ARG__user_id] = user_id

                        if method.parallel_verification:
                            supply_dict[ARG__verification_hook] = verification_hook

                        # Do not close created interfaces as they are re-used
                        for arg in inspect.getfullargspec(view_func_local).args:
                            if arg.startswith(ARG_START__connection):
                                if not swagger_documentation.security:
                                    raise Exception(ERR__method_required_user_connection)
                                connect_db = arg.split(ARG_START__connection)[1]
                                supply_dict[arg] = self.model.create_interface_for_db(user_id, connect_db, sub_role=user_id)

                        for arg in inspect.getfullargspec(view_func_local).args:
                            if arg.startswith(ARG_START__jaaql_connection):
                                connect_db = arg.split(ARG_START__jaaql_connection)[1]
                                supply_dict[arg] = self.model.create_interface_for_db(user_id, connect_db, sub_role=ROLE__jaaql)

                        if ARG__username in inspect.getfullargspec(view_func_local).args:
                            if not swagger_documentation.security:
                                raise Exception(ERR__method_required_username)
                            supply_dict[ARG__username] = username

                        if ARG__ip_address in inspect.getfullargspec(view_func_local).args:
                            supply_dict[ARG__ip_address] = ip_addr

                        if ARG__response in inspect.getfullargspec(view_func_local).args:
                            supply_dict[ARG__response] = jaaql_resp

                        if ARG__auth_token in inspect.getfullargspec(view_func_local).args:
                            if not swagger_documentation.security:
                                raise Exception(ERR__method_required_token)
                            supply_dict[ARG__auth_token] = request.headers.get(HEADER__security)

                        if ARG__connection in inspect.getfullargspec(view_func_local).args:
                            if not swagger_documentation.security:
                                raise Exception(ERR__method_required_connection)
                            supply_dict[ARG__connection] = self.model.create_interface_for_db(user_id, DB__jaaql)

                        if ARG__is_public in inspect.getfullargspec(view_func_local).args:
                            if not swagger_documentation.security:
                                raise Exception(ERR__method_required_is_public)
                            supply_dict[ARG__is_public] = is_public

                        self.perform_profile(request_id, "Fetch args")
                        if ARG__profiler in inspect.getfullargspec(view_func_local).args:
                            supply_dict[ARG__profiler] = Profiler(request_id, self.do_profiling)

                        resp = view_func_local(**supply_dict)

                        self.perform_profile(request_id, "Perform work")

                        if not swagger_documentation.security:
                            user_id = jaaql_resp.user_id
                            ip_id = jaaql_resp.ip_id

                        status = jaaql_resp.response_code
                        method_response = BaseJAAQLController.get_response(the_method, status)
                        do_allow_all = False
                        if len(the_method.responses) != 0:
                            if the_method.responses[0] == RES__allow_all:
                                do_allow_all = True
                        if not do_allow_all:
                            resp = BaseJAAQLController.validate_output(method_response, resp)
                        ret_status = status

                        self.perform_profile(request_id, "Validate output")
                    except Exception as ex:
                        if not self.is_prod:
                            traceback.print_exc()  # Debugging
                        if not isinstance(ex, HttpStatusException):
                            ret_status = RESP__default_err_code
                            ex_msg = RESP__default_err_message
                        else:
                            ret_status = ex.response_code
                            ex_msg = ex.message

                        do_allow_all = False
                        if len(the_method.responses) != 0:
                            if the_method.responses[0] == RES__allow_all:
                                do_allow_all = True

                        if not do_allow_all and ret_status != HTTPStatus.UNAUTHORIZED and ret_status != \
                                HTTPStatus.NOT_IMPLEMENTED and ret_status != HTTPStatus.BAD_REQUEST and \
                                ret_status != HTTPStatus.UNPROCESSABLE_ENTITY and \
                                ret_status != CustomHTTPStatus.DATABASE_NO_EXIST and \
                                ret_status != HTTPStatus.INTERNAL_SERVER_ERROR:
                            try:
                                self.get_response(the_method, ret_status)
                            except Exception as sub_ex:
                                # The expected response code was not allowed
                                traceback.print_exc()
                                ret_status = RESP__default_err_code
                                ex_msg = RESP__default_err_message
                                ex = sub_ex

                        throw_ex = ex

                    duration = round((datetime.now() - start_time).total_seconds() * 1000)
                    if user_id is not None:
                        self.model.log(user_id, start_time, duration, ex_msg, method_input, ip_id, ret_status, route)

                    self.perform_profile(request_id, "Cleanup")

                    if throw_ex is not None:
                        raise throw_ex

                if jaaql_resp.response_type == resp_type:
                    if not isinstance(resp, Response):
                        resp = jsonify(resp)
                    resp.status = jaaql_resp.response_code
                else:
                    resp = Response(resp, mimetype=jaaql_resp.response_type, status=jaaql_resp.response_code)

                self._cors(resp)
                self.perform_profile(request_id, "Jsonify")

                if not BaseJAAQLController.is_options():
                    print("Total time taken: " + str(time_delta_ms(start_time, datetime.now())) + "ms")

                return resp

            self.app.add_url_rule(route, view_func=lambda: routed_function(view_func), methods=methods,
                                  endpoint=route)

            return routed_function

        return wrap_func

    @staticmethod
    def _init_error_handlers(app):
        @app.errorhandler(HTTPStatus.INTERNAL_SERVER_ERROR)
        def handle_server_error(error: InternalServerError):
            if os.environ.get(ENVIRON__sentinel_url):
                orig = error.original_exception

                tb_frame = sys.exc_info()[2]
                while tb_frame.tb_next:
                    tb_frame = tb_frame.tb_next
                source_file = tb_frame.tb_frame.f_code.co_filename[len(os.getcwd()) + 1:]
                source_file = source_file.replace("\\", "/")

                BaseJAAQLController.sentinel_errors.put({
                    "error_condensed": str(orig),
                    "version": VERSION,
                    "source_system": "Sentinel" if BaseJAAQLController.internal_sentinel else "JAAQL",
                    "source_file": source_file,
                    "file_line_number": tb_frame.tb_lineno,
                    "stacktrace": ''.join(traceback.format_exception(etype=type(orig), value=orig, tb=orig.__traceback__))
                })

            traceback.print_tb(error.__traceback__)
            return BaseJAAQLController._cors(Response(RESP__default_err_message, RESP__default_err_code))

        @app.errorhandler(HTTPException)
        def handle_other_server_error(error: HTTPException):
            return BaseJAAQLController._cors(Response(error.description, error.code))

        @app.errorhandler(HttpStatusException)
        def handle_pipeline_exception(error: HttpStatusException):
            if not isinstance(error.response_code, int) or isinstance(error.response_code, CustomHTTPStatus):
                error.response_code = error.response_code.value
            if not isinstance(error.message, str):
                res = jsonify(error.message)
            else:
                res = Response(error.message)
            res.status = error.response_code
            return BaseJAAQLController._cors(res)

    @staticmethod
    def is_post():
        return request.method == REST__POST

    @staticmethod
    def is_get():
        return request.method == REST__GET

    @staticmethod
    def is_delete():
        return request.method == REST__DELETE

    @staticmethod
    def is_put():
        return request.method == REST__PUT

    @staticmethod
    def is_options():
        return request.method == REST__OPTIONS



import os
import traceback
import uuid

from http import HTTPStatus
from jaaql.db.db_interface import DBInterface
from jaaql.db.db_utils import execute_supplied_statement, execute_supplied_statement_singleton
from jaaql.exceptions.http_status_exception import HttpStatusException
from jaaql.utilities.utils import load_config, await_jaaql_installation, get_jaaql_connection, time_delta_ms
from jaaql.utilities.crypt_utils import decrypt_raw
from urllib.parse import quote
from jaaql.utilities.utils_no_project_imports import check_allowable_file_path
import json
import imaplib
import smtplib
import base64
from base64 import urlsafe_b64decode as b64d, urlsafe_b64encode as b64e
from typing import Union, List, Optional
from datetime import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText
from queue import Queue, Empty
import threading
import ssl
import time
import sys
from flask import Flask, jsonify, request
from jaaql.utilities.vault import Vault, DIR__vault
from jaaql.constants import VAULT_KEY__db_crypt_key, KEY__encrypted_password, KEY__id, KEY__template, KEY__sender, ENCODING__ascii, PORT__ems, \
    ENDPOINT__reload_accounts, ENDPOINT__send_email, KEY__attachment_name, KEY__parameters, KEY__url, KEY__oauth_token, \
    KEY__document_id, KEY__create_file, DIR__render_template, KEY__render_as, KEY__content, DIR__www, KEY__email_account_name, \
    KEY__email_account_send_name, KEY__email_account_protocol, KEY__email_account_host, KEY__email_account_port, KEY__email_account_username, \
    KEY__override_send_name, KEY__application, ENVIRON__install_path, KEY__artifact_base_uri

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.options import Options

QUERY__mark_rendered_document_completed = "UPDATE rendered_document SET completed = current_timestamp, filename = :filename, content = :content WHERE document_id = :document_id"
QUERY__fetch_unrendered_document = "SELECT able.url, ac.artifact_base_uri, able.render_as, rd.document_id, rd.encrypted_parameters as parameters, rd.create_file, rd.encrypted_access_token as oauth_token FROM rendered_document rd INNER JOIN renderable_document able ON rd.document = able.name INNER JOIN application_configuration ac ON rd.configuration = ac.name AND rd.application = ac.application WHERE rd.completed is null ORDER BY rd.created LIMIT 1"
QUERY__update_email_account_password = "UPDATE email_account SET encrypted_password = :encrypted_password WHERE id = :id"
QUERY__load_email_accounts = "SELECT * FROM email_account"
QUERY__ins_email_history = "INSERT INTO email_history (application, template, sender, encrypted_subject, encrypted_body, encrypted_attachments, encrypted_recipients, encrypted_recipients_keys, override_send_name) VALUES (:application, :template, :sender, :encrypted_subject, :encrypted_body, :encrypted_attachments, :encrypted_recipients, :encrypted_recipients_keys, :override_send_name)"
QUERY__fetch_email_template = "SELECT rd.url FROM renderable_document rd INNER JOIN renderable_document_template rdt ON rd.name = rdt.attachment WHERE rd.name = :name AND rdt.template = :template"
QUERY__purge_rendered_documents = "DELETE FROM rendered_document rd USING renderable_document able WHERE rd.document = able.name AND completed is not null and completed > current_timestamp + interval '5 minutes' RETURNING rd.document_id, rd.create_file, able.render_as"

KEY__email_account = "email_account"
KEY__encrypted_subject = "encrypted_subject"
KEY__encrypted_body = "encrypted_body"
KEY__encrypted_attachments = "encrypted_attachments"
KEY__encrypted_recipients = "encrypted_recipients"
KEY__encrypted_recipients_keys = "encrypted_recipients_keys"
KEY__artifact_base = "artifact_base"

ATTR__filename = "filename"

ELE__jaaql_filename = "jaaql_filename"

ERR__password_not_found = "Password not found for email account with name '%s'"
ERR__email_not_found = "Email account not found with name '%s'"
ERR__invalid_call_to_internal_email_service = "Invalid call to internal email service"
ERR__attachment_timeout = "Could not send email! Attachment rendering timed out!"
ERR__attachment_error = "Could not send email! Attachment rendering received error: '%s'"
ERR__attachment_filename = "Attachment did not render in time or could not find the file name for the attachment. Please check template. Expected " \
    "presence of element with id " + ELE__jaaql_filename + " to be present after page has finished loading"

PROTOCOL__imap = "imap"
PROTOCOL__smtp = "smtp"

EMAIL__from = "From"
EMAIL__from_email = "From_Email"
EMAIL__to = "To"
EMAIL__subject = "Subject"

SPLIT__address = ", "

TIMEOUT__attachment = 60000
TIMEOUT__filename = 15000


class DrivenChrome:
    def __init__(self, db_interface: DBInterface, db_key: bytes, is_deployed: bool):
        self.attachments = Queue()
        self.completed_attachments = set()
        self.errored_attachments = {}
        self.db_interface = db_interface
        self.chrome_lock = threading.Lock()
        self.driver = None
        self.is_deployed = is_deployed
        self.db_key = db_key
        self.template_dir_path = os.path.join(DIR__www, DIR__render_template)

        self.a4_params = {'landscape': False, 'paperWidth': 8.27, 'paperHeight': 11.69, 'printBackground': True}

        threading.Thread(target=self.start_chrome, daemon=True).start()
        threading.Thread(target=self.purge_rendered_documents, daemon=True).start()

    def parameters_to_get_str(self, access_token: str, parameters: dict):
        if parameters is None:
            parameters = {}
        parameters[KEY__oauth_token] = access_token

        return "?" + "&".join([key + "=" + quote(itm) for key, itm in parameters.items()])

    def chrome_page_to_pdf(self, url: str, access_token: str, parameters: dict):
        filename = None

        with self.chrome_lock:
            self.driver.get(url + self.parameters_to_get_str(access_token, parameters))
            start_time = datetime.now()
            while True:
                filename = self.driver.find_elements(By.ID, ELE__jaaql_filename)
                if len(filename) != 0:
                    file_text = filename[0].get_attribute("innerHTML")
                    if not file_text.endswith(".pdf"):
                        file_text += ".pdf"
                    filename = file_text
                    break
                else:
                    if time_delta_ms(start_time, datetime.now()) > TIMEOUT__filename:
                        raise HttpStatusException(ERR__attachment_filename)
                    time.sleep(0.25)

            return filename, base64.b64decode(self.driver.execute_cdp_cmd("Page.printToPDF", self.a4_params)["data"])

    def purge_rendered_documents(self):
        while True:
            resp = execute_supplied_statement(self.db_interface, QUERY__purge_rendered_documents, as_objects=True)
            for to_purge in resp:
                if to_purge[KEY__create_file]:
                    try:
                        os.remove(os.path.join(self.template_dir_path, str(to_purge[KEY__document_id]) + "." + to_purge[KEY__render_as]))
                    except FileNotFoundError:
                        pass
            time.sleep(5)

    def render_document_requests(self):
        while True:
            try:
                resp = execute_supplied_statement_singleton(self.db_interface, QUERY__fetch_unrendered_document, as_objects=True,
                                                            decrypt_columns=[KEY__parameters, KEY__oauth_token], encryption_key=self.db_key)
                parameters = {}
                if resp[KEY__parameters]:
                    parameters = json.loads(resp[KEY__parameters])

                base_uri = EmailAttachment.static_format_attached_url(resp[KEY__artifact_base_uri], self.is_deployed)
                filename, content = self.chrome_page_to_pdf(base_uri, resp[KEY__oauth_token], parameters)

                inputs = {KEY__document_id: resp[KEY__document_id], KEY__content: None, ATTR__filename: filename}
                if resp[KEY__create_file]:
                    if not os.path.exists(self.template_dir_path):
                        os.mkdir(self.template_dir_path)
                    with open(os.path.join(self.template_dir_path, str(resp[KEY__document_id]) + "." + resp[KEY__render_as]), "wb") as f:
                        f.write(content)
                else:
                    inputs[KEY__content] = content

                execute_supplied_statement(self.db_interface, QUERY__mark_rendered_document_completed, inputs)
            except HttpStatusException as ex:
                if ex.response_code == HTTPStatus.UNPROCESSABLE_ENTITY:
                    time.sleep(0.25)
                else:
                    traceback.print_exc()

    def start_chrome(self):
        options = Options()
        options.add_argument("--window-size=1920,1080")
        options.headless = True
        service_log_path = None

        if self.is_deployed:
            options.add_argument('--no-sandbox')
            service_log_path = "log/chromedriver.log"

        with webdriver.Chrome(options=options, service_log_path=service_log_path) as driver:
            self.driver = driver
            threading.Thread(target=self.render_document_requests, daemon=True).start()
            while True:
                current_attachment: 'EmailAttachment' = self.attachments.get()
                try:
                    url = execute_supplied_statement_singleton(self.db_interface, QUERY__fetch_email_template,
                                                               {KEY__attachment_name: current_attachment.name,
                                                                KEY__template: current_attachment.template}, as_objects=True)[KEY__url]
                    if not url.endswith(".html"):
                        url += ".html"
                    url = current_attachment.artifact_base + "/" + url
                    filename, content = self.chrome_page_to_pdf(url, current_attachment.access_token, current_attachment.parameters)
                    current_attachment.content = content
                    current_attachment.filename = filename
                except HttpStatusException as ex:
                    self.errored_attachments[current_attachment.id] = str(ex)
                    self.completed_attachments.add(current_attachment.id)
                except Exception as ex:
                    traceback.print_exc()
                    self.errored_attachments[current_attachment.id] = str(ex)
                    self.completed_attachments.add(current_attachment.id)
                self.completed_attachments.add(current_attachment.id)


class EmailAttachment:
    def __init__(self, name: str, parameters: dict, template: str, artifact_base: str = None):
        self.name = name
        self.artifact_base = artifact_base
        self.parameters = parameters
        self.id = str(uuid.uuid4())
        self.filename = None
        self.content = None
        self.access_token = None
        self.template = str(template)

    @staticmethod
    def static_format_attached_url(artifact_base_uri: str, is_container: bool):
        if not artifact_base_uri.startswith("https://") and not artifact_base_uri.startswith("http://"):
            if is_container:
                if check_allowable_file_path(artifact_base_uri):
                    raise Exception("Illegal artifact base url")

            if artifact_base_uri.startswith("file:///"):
                return artifact_base_uri
            else:
                return "file://" + os.environ[ENVIRON__install_path] + "/www/" + artifact_base_uri
        else:
            return artifact_base_uri

    def format_attachment_url(self, artifact_base_uri: str, is_container: bool):
        self.artifact_base = EmailAttachment.static_format_attached_url(artifact_base_uri, is_container)

    def build_attachment(self, access_token: str, driven_chrome: DrivenChrome) -> MIMEApplication:
        self.access_token = access_token
        driven_chrome.attachments.put(self)

        start_time = datetime.now()
        while time_delta_ms(start_time, datetime.now()) < TIMEOUT__attachment:
            if self.id in driven_chrome.completed_attachments:
                break
            time.sleep(1)

        if self.id not in driven_chrome.completed_attachments:
            raise Exception(ERR__attachment_timeout)
        driven_chrome.completed_attachments.remove(self.id)

        if self.id in driven_chrome.errored_attachments:
            raise Exception(ERR__attachment_error % driven_chrome.errored_attachments.pop(self.id))

        attachment = MIMEApplication(self.content, _subtype=self.filename.split(".")[-1])
        attachment.add_header('Content-Disposition', 'attachment', filename=self.filename)

        return attachment

    def repr_json(self):
        return dict(name=self.name, parameters=self.parameters, template=self.template, artifact_base=self.artifact_base)

    @staticmethod
    def deserialize(attachment: dict, template: str = None, artiface_base_url: str = None):
        if template is not None:
            return EmailAttachment(attachment[KEY__attachment_name], attachment[KEY__parameters], template, artiface_base_url)
        else:
            return EmailAttachment(attachment[KEY__attachment_name], attachment[KEY__parameters], attachment[KEY__template],
                                   attachment[KEY__artifact_base])

    @staticmethod
    def deserialize_list(attachments: list, template: str, artiface_base_url: str):
        if attachments is None:
            return None
        else:
            return [EmailAttachment.deserialize(attachment, template, artiface_base_url) for attachment in attachments]

    def encode_filename(self):
        return b64e(self.filename.encode(ENCODING__ascii)).decode(ENCODING__ascii)

    def encode_content(self):
        return b64e(self.content).decode(ENCODING__ascii)


TYPE__email_attachments = Union[EmailAttachment, List[EmailAttachment]]


class Email:
    def __init__(self, sender: str, application: str, template: str, from_account: str, to: Union[str, List[str]],
                 recipient_names: Union[str, List[str]], subject: str = None, body: str = None, attachments: TYPE__email_attachments = None,
                 is_html: bool = True, attachment_access_token: str = None, override_send_name: str = None):
        self.sender = sender
        self.application = application
        self.template = template
        self.from_account = from_account
        if isinstance(to, list) != isinstance(recipient_names, list):
            raise Exception(ERR__invalid_call_to_internal_email_service)
        if not isinstance(to, list):
            to = [to]
            recipient_names = [recipient_names]
        if len(to) != len(recipient_names):
            raise Exception(ERR__invalid_call_to_internal_email_service)
        recipient_names = [to[idx] if itm is None else itm for idx, itm in zip(range(len(recipient_names)), recipient_names)]
        self.to = to
        self.recipient_names = recipient_names
        self.subject = subject
        self.body = body
        self.attachments = attachments
        self.override_send_name = override_send_name
        if not isinstance(self.attachments, List) and self.attachments is not None:
            self.attachments = [attachments]
        self.is_html = is_html
        self.attachment_access_token = attachment_access_token

    def repr_json(self):
        return dict(sender=self.sender, application=self.application, template=self.template, from_account=self.from_account,
                    to=self.to, recipient_names=self.recipient_names, subject=self.subject, body=self.body,
                    attachments=[attachment.repr_json() for attachment in self.attachments] if self.attachments is not None else None,
                    is_html=self.is_html, attachment_access_token=self.attachment_access_token, override_send_name=self.override_send_name)

    @staticmethod
    def deserialize(email: dict):
        attachments = None
        if email["attachments"] is not None:
            attachments = [EmailAttachment.deserialize(attachment) for attachment in email["attachments"]]
        return Email(email["sender"], email["application"], email["template"], email["from_account"], email["to"],
                     email["recipient_names"], email["subject"], email["body"], attachments, email["is_html"], email["attachment_access_token"],
                     email["override_send_name"])


class EmailManagerService:

    def __init__(self, connection: DBInterface, email_credentials: Optional[str], db_crypt_key: bytes, is_gunicorn: bool):
        if email_credentials is None:
            self.email_credentials = {}
        else:
            self.email_credentials = json.loads(b64d(email_credentials).decode(ENCODING__ascii))
        self.db_crypt_key = db_crypt_key
        self.connection = connection

        self.email_queues = {}

        self.is_gunicorn = is_gunicorn

        self.reload_lock = threading.Lock()
        self.reload_accounts()

        self.driven_chrome = DrivenChrome(connection, self.db_crypt_key, self.is_gunicorn)

    def reload_accounts(self):
        with self.reload_lock:
            accounts = execute_supplied_statement(self.connection, QUERY__load_email_accounts, as_objects=True)
            for account in accounts:
                account[KEY__encrypted_password] = decrypt_raw(self.db_crypt_key, account[KEY__encrypted_password])
            email_queues = {}

            for account in accounts:
                if account[KEY__email_account_name] not in self.email_credentials.credentials and account[KEY__encrypted_password] is None:
                    raise Exception(ERR__password_not_found % account[KEY__email_account_name])
                elif account[KEY__email_account_name] in self.email_credentials.credentials and account[KEY__encrypted_password] is None:
                    execute_supplied_statement(self.connection, QUERY__update_email_account_password,  # Stores the password
                                               {KEY__encrypted_password: self.email_credentials.credentials[account[KEY__email_account_name]],
                                                KEY__id: account[KEY__id]},
                                               encrypt_parameters=[KEY__encrypted_password],
                                               encryption_key=self.db_crypt_key)
                cur_queue = Queue()

                if str(account[KEY__email_account_name]) not in self.email_queues:
                    email_queues[str(account[KEY__email_account_name])] = cur_queue
                    password = account[KEY__encrypted_password]
                    if password is None:
                        password = self.email_credentials[account[KEY__email_account_name]]
                    method_args = [account, cur_queue, password]
                    threading.Thread(target=self.email_connection_thread, args=method_args, daemon=True).start()
                else:
                    email_queues[str(account[KEY__email_account_name])] = self.email_queues[str(account[KEY__email_account_name])]

            self.email_queues = email_queues

    def construct_message(self, account: dict, email: Email) -> (str, MIMEMultipart):
        send_body = email.body

        message = MIMEMultipart("alternative")

        if email.is_html:
            message.attach(MIMEText(send_body, 'html'))
        else:
            message.attach(MIMEText(send_body, 'plain'))

        message_final = MIMEMultipart('mixed')
        message_final.attach(message)
        message_final[EMAIL__to] = SPLIT__address.join(email.to)
        send_name = account[KEY__email_account_send_name]
        if email.override_send_name is not None:
            send_name = email.override_send_name
        message_final[EMAIL__from] = send_name + " <" + account[KEY__email_account_username] + ">"
        message_final[EMAIL__subject] = email.subject

        if email.attachments is not None:
            email_attachments = email.attachments
            for attachment in email_attachments:
                message_final.attach(attachment.build_attachment(email.attachment_access_token, self.driven_chrome))

        return send_body, message_final

    def is_connected(self, conn: smtplib.SMTP):
        try:
            status = conn.noop()[0]
        except Exception:
            status = -1
        return True if status == 250 else False

    def fetch_conn(self, conn_lib, account: dict, password: str):
        conn = conn_lib(account[KEY__email_account_host], port=account[KEY__email_account_port])

        context = ssl.SSLContext(ssl.PROTOCOL_TLS)

        if conn_lib == smtplib.SMTP_SSL:
            conn.ehlo()
        conn.starttls(context=context)
        if conn_lib == smtplib.SMTP_SSL:
            conn.ehlo()
        conn.login(account[KEY__email_account_username], password)

        return conn

    def format_attachments_for_storage(self, attachments: TYPE__email_attachments):
        if attachments is None:
            return None
        if not isinstance(attachments, list):
            attachments = [attachments]
        attachments = "::".join([attachment.encode_filename() + ":" + attachment.encode_content() for attachment in attachments])
        return attachments

    def email_connection_thread(self, account: dict, queue: Queue, password: str):
        conn_lib = None
        conn = None
        if self.is_gunicorn:
            conn_lib = imaplib.IMAP4 if account[KEY__email_account_protocol] == PROTOCOL__imap else smtplib.SMTP
            conn = self.fetch_conn(conn_lib, account, password)

        while True:
            try:
                email: Email = queue.get(timeout=10)
                if self.is_gunicorn:
                    if not self.is_connected(conn):
                        conn = self.fetch_conn(conn_lib, account, password)
                formatted_body, to_send = self.construct_message(account, email)
                if self.is_gunicorn:
                    conn.send_message(to_send, account[KEY__email_account_username], email.to)
                else:
                    print("SENDING EMAIL")
                    print(email.subject)
                    print(email.body)
                    print(account[KEY__email_account_username])
                    print(email.to)

                execute_supplied_statement(self.connection, QUERY__ins_email_history, {
                        KEY__application: email.application,
                        KEY__template: email.template,
                        KEY__sender: email.sender,
                        KEY__override_send_name: email.override_send_name,
                        KEY__encrypted_subject: email.subject,
                        KEY__encrypted_body: formatted_body,
                        KEY__encrypted_attachments: self.format_attachments_for_storage(email.attachments),
                        KEY__encrypted_recipients: SPLIT__address.join(email.to),
                        KEY__encrypted_recipients_keys: SPLIT__address.join(email.recipient_names)
                    }, encryption_key=self.db_crypt_key, encrypt_parameters=[
                        KEY__encrypted_subject,
                        KEY__encrypted_body,
                        KEY__encrypted_attachments,
                        KEY__encrypted_recipients,
                        KEY__encrypted_recipients_keys
                    ]
                )
            except Empty:
                if account[KEY__email_account_name] not in self.email_queues:
                    break
            except Exception:
                traceback.print_exc()  # Something went wrong

        try:
            conn.quit()
        except Exception:
            pass

    def send_email(self, email: Email):
        if email.from_account in self.email_queues:
            self.email_queues[email.from_account].put(email)
        else:
            self.reload_accounts()
            if email.from_account in self.email_queues:
                self.email_queues[email.from_account].put(email)
            else:
                raise Exception(ERR__email_not_found % email.from_account)


def create_app(ems: EmailManagerService):

    app = Flask(__name__, instance_relative_config=True)

    @app.route(ENDPOINT__send_email, methods=["POST"])
    def send_email():
        ems.send_email(Email.deserialize(request.json))

        return jsonify("OK")

    @app.route(ENDPOINT__reload_accounts, methods=["POST"])
    def refresh_accounts():
        ems.reload_accounts()

        return jsonify("OK")

    @app.route("/", methods=["GET"])
    def is_alive():
        return jsonify("OK")

    return app


def create_flask_app(vault_key=None, email_credentials=None, is_gunicorn: bool = False):
    config = load_config(is_gunicorn)
    await_jaaql_installation(config, is_gunicorn)
    vault = Vault(vault_key, DIR__vault)
    jaaql_lookup_connection = get_jaaql_connection(config, vault)
    db_crypt_key = vault.get_obj(VAULT_KEY__db_crypt_key).encode(ENCODING__ascii)

    flask_app = create_app(EmailManagerService(jaaql_lookup_connection, email_credentials, db_crypt_key, is_gunicorn))
    print("Created email manager app host, running flask", file=sys.stderr)
    flask_app.run(port=PORT__ems, host="0.0.0.0", threaded=True)



from jaaql.utilities.utils import load_config, await_ems_startup, get_base_url, await_jaaql_bootup
from jaaql.constants import HEADER__security_bypass, VAULT_KEY__jaaql_local_access_key, ENDPOINT__install, ENVIRON__install_path
from jaaql.utilities.vault import Vault, DIR__vault
import os
from os.path import join, exists, dirname
from os import listdir
import requests
import json
import time

DIR__install_scripts = "install_scripts"


def process_script(script: str, base_url: str, bypass_header: dict):
    script = script.replace("\r\n", "\n")
    commands = script.split("\n\n")
    if len(commands[0]) == 0:
        print("Script is empty")
    for command in commands:
        action = command.split("\n")[0].strip()
        method = action.split(" ")[0]
        endpoint = action.split(" ")[1]

        loaded_json = None
        if len(command.split("\n")) > 1:
            loaded_json = json.loads("\n".join(command.split("\n")[1:]).strip())

        requests.request(method, base_url + endpoint, headers=bypass_header, json=loaded_json)


def bootup(vault_key, is_gunicorn: bool = False, install_on_bootup: bool = False):
    config = load_config(is_gunicorn)
    install_status = await_jaaql_bootup(config, is_gunicorn)
    base_url = get_base_url(config, is_gunicorn)
    if install_on_bootup and install_status != 200:
        install_key_file = join(dirname(dirname(dirname(__file__))), "install_key")
        while not exists(install_key_file):
            time.sleep(0.1)
        time.sleep(0.5)
        install_key = None
        with open(install_key_file, "r") as install_key_file:
            install_key = install_key_file.read().strip()
        json_data = {
            "super_db_password": "pa55word",
            "jaaql_password": "pa55word",
            "install_key": install_key,
            "allow_uninstall": False
        }
        if not is_gunicorn:
            json_data["db_connection_string"] = "postgresql://postgres:123456@localhost:5434/"
        requests.post(base_url + ENDPOINT__install, json=json_data)

    await_ems_startup()
    vault = Vault(vault_key, DIR__vault)

    bypass_header = {HEADER__security_bypass: vault.get_obj(VAULT_KEY__jaaql_local_access_key)}
    if ENVIRON__install_path in os.environ:
        did_script_install_marker = join(os.environ[ENVIRON__install_path], "vault", "did_script_install")
    else:
        did_script_install_marker = join("vault", "did_script_install")
    try:
        os.makedirs(did_script_install_marker)
    except FileExistsError:
        pass
    script_root = join(os.getcwd(), DIR__install_scripts)
    try:
        i = 0
        for script in listdir(script_root):
            cur_script_marker = join(did_script_install_marker, script)
            if not exists(cur_script_marker):
                if i != 0:
                    print()

                with open(join(script_root, script), "r") as script_file:
                    print("Processing script " + script)
                    process_script(script_file.read(), base_url, bypass_header)
                i += 1
            open(cur_script_marker, "w").close()
    except FileNotFoundError:
        print("Skipping installation as no files found in install scripts")



from jaaql.db.db_interface import DBInterface
from jaaql.utilities.utils import time_delta_ms, get_jaaql_root, loc, get_base_url, get_external_url
from datetime import datetime
import hashlib
from os.path import join, isdir
from os import listdir, environ
from jaaql.interpreter.interpret_jaaql import InterpretJAAQL
from jaaql.exceptions.http_status_exception import HttpStatusException
from jaaql.db.db_utils import create_interface
from jaaql.utilities.crypt_utils import encrypt_raw
import re
from jaaql.utilities.crypt_utils import AES__iv_length

MIGRATION_HISTORY = "migration_history"


QUERY_LOAD_TABLE = "SELECT * FROM %s WHERE project_name = :project_name" % MIGRATION_HISTORY
ATTR_PROJECT_NAME = "project_name"
QUERY_INS_TABLE = "INSERT INTO %s (project_name, installed_rank, version, description, script, checksum, execution_time) VALUES (:project_name, :installed_rank, :version, :description, :script, :checksum, :execution_time)" % MIGRATION_HISTORY
ATTR_INSTALLED_RANK = "installed_rank"
ATTR_VERSION = "version"
ATTR_DESCRIPTION = "description"
ATTR_SCRIPT = "script"
ATTR_CHECKSUM = "checksum"
ATTR_EXECUTION_TIME = "execution_time"
PATH_MIGRATIONS = "migrations"
SCRIPT_MIGRATION_HISTORY = "migration_history.sql"
EXTENSION_SQL = ".sql"

VERSION_SPLIT = "__"
WORD_SPLIT = "_"

PROJECT_JAAQL = "JAAQL"

REGEX__migration_replacement = r'{{JAAQL__[A-Z_]+}}'
REGEX__migration_replacement_and_encrypt = r'#{{JAAQL__[A-Z_]+}}'


def load_config_file(location: str) -> (str, str, str):
    ret_role = None
    ret_databases = None
    try:
        the_content = open(location, "r").read()

        if "ROLE=" in the_content:
            ret_role = the_content.split("ROLE=")[1].split("\r")[0].split("\n")[0].strip()
        if "DATABASES=" in the_content:
            ret_databases = the_content.split("DATABASES=")[1].split("\r")[0].split("\n")[0].strip()
    except:
        pass

    return ret_role, ret_databases


def get_config(role: str, folder_role: str, file_role: str, databases: str,
               folder_databases: str, file_databases: str):
    ret_role = "jaaql"
    ret_databases = ["jaaql"]

    if role is not None:
        ret_role = role
    if folder_role is not None:
        ret_role = folder_role
    if file_role is not None:
        ret_role = file_role

    if databases is not None:
        ret_databases = databases.split(",")
    if folder_databases is not None:
        ret_databases = folder_databases.split(",")
    if file_databases is not None:
        ret_databases = file_databases.split(",")

    return ret_role, ret_databases


def create_interface_for_db(config, super_credentials: str, database: str, sub_role: str):
    address, port, _, username, password = DBInterface.fracture_uri(super_credentials)

    return create_interface(config, address, port, database, username, password=password, sub_role=sub_role)


def replace_options_environment_variables(query: str, parsed_options: dict, crypt_key: bytes, is_deployed: bool, config, fixed_salt: bool):
    last_index = 0
    prepared = ""
    for match in re.finditer(REGEX__migration_replacement_and_encrypt, query):
        start_pos = match.regs[0][0]
        end_pos = match.regs[0][1]
        match_str = query[start_pos + 3 + len("JAAQL__"):end_pos - 2]
        prepared += query[last_index:start_pos]
        last_index = end_pos

        option_match_str = match_str.lower().replace("_", "-")
        if option_match_str in parsed_options:
            to_crypt = parsed_options[option_match_str]
        elif match_str in environ:
            to_crypt = environ[match_str]
        else:
            raise Exception("Missing parameter from migration scripts " + match_str)
        iv = None
        if fixed_salt:
            iv = b"\x00" * AES__iv_length
        prepared += encrypt_raw(crypt_key, to_crypt, iv=iv)

    prepared += query[last_index:]
    query = prepared

    last_index = 0
    prepared = ""
    for match in re.finditer(REGEX__migration_replacement, query):
        start_pos = match.regs[0][0]
        end_pos = match.regs[0][1]
        match_str = query[start_pos + 2 + len("JAAQL__"):end_pos - 2]
        is_quote = match_str.startswith("QUOTE_")
        if is_quote:
            match_str = match_str[len("QUOTE_"):]
        is_double_quote = match_str.startswith("QUOTE_")
        if is_double_quote:
            match_str = match_str[len("DOUBLE_QUOTE_"):]
        prepared += query[last_index:start_pos]
        last_index = end_pos

        option_match_str = match_str.lower().replace("_", "-")
        was_null = False
        if option_match_str in parsed_options:
            to_add = parsed_options[option_match_str]
        elif match_str in environ:
            to_add = environ[match_str]
        elif match_str == "BASE_URL":
            to_add = get_base_url(config, is_deployed)
        elif match_str == "EXTERNAL_URL":
            to_add = get_external_url(config)
        else:
            was_null = True
            to_add = "null"

        if not was_null and is_quote:
            prepared += "'"
        if not was_null and is_double_quote:
            prepared += '"'
        prepared += to_add
        if not was_null and is_quote:
            prepared += "'"
        if not was_null and is_double_quote:
            prepared += '"'

    prepared += query[last_index:]
    return prepared


def run_migrations(db_interface: DBInterface, is_deployed: bool, project_name: str = None, migration_folder: str = None, config=None,
                   super_credentials: str = None, options: dict = None, key: bytes = None):
    if migration_folder is None:
        migration_folder = join(get_jaaql_root(), PATH_MIGRATIONS)

    base_override_role, base_override_databases = load_config_file(join(migration_folder, "config"))

    if project_name is None:
        project_name = PROJECT_JAAQL

    ij = InterpretJAAQL(db_interface)

    migration_history = None
    conn = db_interface.get_conn()

    statement_load_table = {"query": QUERY_LOAD_TABLE, "parameters": {ATTR_PROJECT_NAME: project_name}}

    dbs = {}

    try:
        migration_history = db_interface.objectify(ij.transform(statement_load_table, conn=conn))
    except HttpStatusException:
        print("Migration history table does not exist. Creating")
        db_interface.execute_script_file(conn, join(get_jaaql_root(), PATH_MIGRATIONS, SCRIPT_MIGRATION_HISTORY))
        migration_history = db_interface.objectify(ij.transform(statement_load_table, conn=conn))

    installed_scripts = [cur[ATTR_SCRIPT] for cur in migration_history]
    cur_installed_rank = ([-1] + sorted([cur[ATTR_INSTALLED_RANK] for cur in migration_history]))[-1]

    version_folders = sorted([version_folder for version_folder in listdir(migration_folder) if
                              isdir(join(migration_folder, version_folder)) and '.' in version_folder])
    for version_folder in version_folders:
        folder_override_role, folder_override_databases = load_config_file(join(migration_folder, version_folder, "config"))

        script_files = sorted([script_file for script_file in listdir(join(migration_folder, version_folder)) if
                               script_file.endswith(EXTENSION_SQL)])
        for script_file in script_files:
            config_file_name = join(migration_folder, version_folder, script_file.split(EXTENSION_SQL)[0] + ".config")
            script_override_role, script_override_databases = load_config_file(config_file_name)
            full_name = version_folder + "/" + script_file
            content = open(join(migration_folder, version_folder, script_file), "r").read()
            cur_role, cur_databases = get_config(base_override_role, folder_override_role, script_override_role,
                                                             base_override_databases, folder_override_databases, script_override_databases)
            content_for_hash = replace_options_environment_variables(content, options, key, is_deployed, config,
                                                                     fixed_salt=True)
            content = replace_options_environment_variables(content, options, key, is_deployed, config,
                                                            fixed_salt=False)

            hash_content = content_for_hash + cur_role + str(cur_databases)
            checksum = hashlib.md5(hash_content.encode("UTF-8")).digest()
            checksum = int.from_bytes(checksum[0:3], byteorder="little")
            if full_name not in installed_scripts:
                start_time = datetime.now()

                for db in cur_databases:
                    if db not in dbs:
                        dbs[db] = create_interface_for_db(config, super_credentials, db, cur_role)
                    update_db_interface = dbs[db]
                    if not hasattr(update_db_interface, "attached_conns"):
                        update_db_interface.attached_conns = {}
                    update_db_interface.sub_role = cur_role
                    if cur_role not in update_db_interface.attached_conns:
                        update_db_interface.attached_conns[cur_role] = update_db_interface.get_conn()
                    update_db_interface.execute_script_file(update_db_interface.attached_conns[cur_role], as_content=content)
                execution_time = time_delta_ms(start_time, datetime.now())
                version = script_file.split("__")[0][1:]
                description = " ".join(script_file.split(VERSION_SPLIT)[1].split(EXTENSION_SQL)[0].split(WORD_SPLIT))
                ij.transform({
                    "query": QUERY_INS_TABLE,
                    "parameters": {
                        ATTR_PROJECT_NAME: project_name,
                        ATTR_INSTALLED_RANK: cur_installed_rank + 1,
                        ATTR_VERSION: version,
                        ATTR_DESCRIPTION: description,
                        ATTR_SCRIPT: full_name,
                        ATTR_CHECKSUM: checksum,
                        ATTR_EXECUTION_TIME: execution_time
                    }
                }, conn=conn)
                cur_installed_rank += 1
            else:
                existing_checksum = loc(migration_history, ATTR_SCRIPT, full_name)
                existing_checksum = loc(existing_checksum, ATTR_PROJECT_NAME, project_name)[0][ATTR_CHECKSUM]
                if checksum != existing_checksum:
                    raise Exception("Migration mismatch for " + script_file + ". Locally calculated checksum " + str(
                        checksum) + " yet in db table found " + str(existing_checksum))

    for cur_db in dbs.values():
        for cur_conn in cur_db.attached_conns.values():
            cur_db.put_conn(cur_conn)
        cur_db.close()

    db_interface.put_conn(conn)
